# Copyright 2017 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import logging

import mock
from neutron.api import extensions as api_ext
from neutron.common import config
import neutron.extensions as nextensions
from oslo_config import cfg
from oslo_utils import importutils
from oslo_utils import uuidutils
import webob.exc

from networking_sfc.db import flowclassifier_db as fdb
from networking_sfc.db import sfc_db
from networking_sfc import extensions
from networking_sfc.extensions import flowclassifier as fc_ext
from networking_sfc.extensions import servicegraph as sg_ext
from networking_sfc.extensions import sfc
from networking_sfc.tests import base
from networking_sfc.tests.unit.db import test_flowclassifier_db


DB_SFC_PLUGIN_CLASS = (
    "networking_sfc.db.sfc_db.SfcDbPlugin"
)
extensions_path = ':'.join(extensions.__path__ + nextensions.__path__)


class SfcDbPluginTestCaseBase(
    base.BaseTestCase
):
    def _assert_port_chain_equal(self, res_port_chain, expected):
        # Flow classifiers are stored in a list, only check items for them
        for k, v in expected.items():
            if type(v) is list:
                self.assertItemsEqual(res_port_chain[k], v)
            else:
                self.assertEqual(res_port_chain[k], v)

    def _create_port_chain(
        self, fmt, port_chain=None, expected_res_status=None, **kwargs
    ):
        ctx = kwargs.get('context', None)
        tenant_id = kwargs.get('tenant_id', self._tenant_id)
        data = {'port_chain': port_chain or {}}
        if ctx is None:
            data['port_chain'].update({'tenant_id': tenant_id})
        req = self.new_create_request(
            'port_chains', data, fmt, context=ctx
        )
        res = req.get_response(self.ext_api)
        if expected_res_status:
            self.assertEqual(expected_res_status, res.status_int)
        return res

    @contextlib.contextmanager
    def port_chain(self, fmt=None, port_chain=None, do_delete=True, **kwargs):
        if not fmt:
            fmt = self.fmt
        res = self._create_port_chain(fmt, port_chain, **kwargs)
        if res.status_int >= 400:
            logging.error('create port chain result: %s', res)
            raise webob.exc.HTTPClientError(code=res.status_int)
        port_chain = self.deserialize(fmt or self.fmt, res)
        yield port_chain
        if do_delete:
            self._delete('port_chains', port_chain['port_chain']['id'])

    def _create_port_pair_group(
        self, fmt, port_pair_group=None, expected_res_status=None, **kwargs
    ):
        ctx = kwargs.get('context', None)
        tenant_id = kwargs.get('tenant_id', self._tenant_id)
        data = {'port_pair_group': port_pair_group or {}}
        if ctx is None:
            data['port_pair_group'].update({'tenant_id': tenant_id})
        req = self.new_create_request(
            'port_pair_groups', data, fmt, context=ctx
        )
        res = req.get_response(self.ext_api)
        if expected_res_status:
            self.assertEqual(expected_res_status, res.status_int)
        return res

    @contextlib.contextmanager
    def port_pair_group(
        self, fmt=None, port_pair_group=None, do_delete=True, **kwargs
    ):
        if not fmt:
            fmt = self.fmt
        res = self._create_port_pair_group(fmt, port_pair_group, **kwargs)
        if res.status_int >= 400:
            logging.error('create port pair group result: %s', res)
            raise webob.exc.HTTPClientError(code=res.status_int)
        port_pair_group = self.deserialize(fmt or self.fmt, res)
        yield port_pair_group
        if do_delete:
            self._delete(
                'port_pair_groups',
                port_pair_group['port_pair_group']['id'])

    def _create_port_pair(
        self, fmt, port_pair=None, expected_res_status=None, **kwargs
    ):
        ctx = kwargs.get('context', None)
        tenant_id = kwargs.get('tenant_id', self._tenant_id)
        data = {'port_pair': port_pair or {}}
        if ctx is None:
            data['port_pair'].update({'tenant_id': tenant_id})
        req = self.new_create_request(
            'port_pairs', data, fmt, context=ctx
        )
        res = req.get_response(self.ext_api)
        if expected_res_status:
            self.assertEqual(expected_res_status, res.status_int)
        return res

    @contextlib.contextmanager
    def port_pair(self, fmt=None, port_pair=None, do_delete=True, **kwargs):
        if not fmt:
            fmt = self.fmt
        res = self._create_port_pair(fmt, port_pair, **kwargs)
        if res.status_int >= 400:
            logging.error('create port pair result: %s', res)
            raise webob.exc.HTTPClientError(code=res.status_int)
        port_pair = self.deserialize(fmt or self.fmt, res)
        yield port_pair
        if do_delete:
            self._delete('port_pairs', port_pair['port_pair']['id'])

    def _create_service_graph(
        self, fmt, service_graph=None, expected_res_status=None, **kwargs
    ):
        ctx = kwargs.get('context', None)
        project_id = kwargs.get('project_id', self._tenant_id)
        data = {'service_graph': service_graph or {}}
        if ctx is None:
            data['service_graph'].update({'project_id': project_id})

        req = self.new_create_request(
            'service_graphs', data, fmt, context=ctx
        )
        res = req.get_response(self.ext_api)
        if expected_res_status:
            self.assertEqual(expected_res_status, res.status_int)
        return res

    @contextlib.contextmanager
    def service_graph(self, fmt=None,
                      service_graph=None, do_delete=True, **kwargs):
        if not fmt:
            fmt = self.fmt

        res = self._create_service_graph(fmt, service_graph, **kwargs)
        if res.status_int >= 400:
            logging.error('create Service Graph result: %s', res)
            raise webob.exc.HTTPClientError(code=res.status_int)
        service_graph = self.deserialize(fmt or self.fmt, res)
        yield service_graph
        if do_delete:
            self._delete('service_graphs', service_graph[
                'service_graph']['id'])

    def _get_expected_port_pair(self, port_pair):
        return {
            'name': port_pair.get('name') or '',
            'description': port_pair.get('description') or '',
            'egress': port_pair.get('egress'),
            'ingress': port_pair.get('ingress'),
            'service_function_parameters': port_pair.get(
                'service_function_parameters') or {
                'correlation': None, 'weight': 1
            }
        }

    def _test_create_port_pair(self, port_pair, expected_port_pair=None):
        if expected_port_pair is None:
            expected_port_pair = self._get_expected_port_pair(port_pair)
        with self.port_pair(port_pair=port_pair) as pp:
            for k, v in expected_port_pair.items():
                self.assertEqual(pp['port_pair'][k], v)

    def _test_create_port_pairs(
        self, port_pairs, expected_port_pairs=None
    ):
        if port_pairs:
            port_pair = port_pairs.pop()
            if expected_port_pairs:
                expected_port_pair = expected_port_pairs.pop()
            else:
                expected_port_pair = self._get_expected_port_pair(port_pair)
            with self.port_pair(port_pair=port_pair) as pp:
                for k, v in expected_port_pair.items():
                    self.assertEqual(pp['port_pair'][k], v)

    def _get_expected_port_pair_group(self, port_pair_group):
        ret = {
            'name': port_pair_group.get('name') or '',
            'description': port_pair_group.get('description') or '',
            'port_pairs': port_pair_group.get('port_pairs') or [],
            'port_pair_group_parameters': port_pair_group.get(
                'port_pair_group_parameters'
            ) or {'lb_fields': [],
                  'ppg_n_tuple_mapping': {'ingress_n_tuple': {},
                                          'egress_n_tuple': {}}}
        }
        if port_pair_group.get('group_id'):
            ret['group_id'] = port_pair_group['group_id']
        return ret

    def _test_create_port_pair_group(
        self, port_pair_group, expected_port_pair_group=None
    ):
        if expected_port_pair_group is None:
            expected_port_pair_group = self._get_expected_port_pair_group(
                port_pair_group)
        with self.port_pair_group(port_pair_group=port_pair_group) as pg:
            for k, v in expected_port_pair_group.items():
                self.assertEqual(pg['port_pair_group'][k], v)

    def _test_create_port_pair_groups(
        self, port_pair_groups, expected_port_pair_groups=None
    ):
        if port_pair_groups:
            port_pair_group = port_pair_groups.pop()
            if expected_port_pair_groups:
                expected_port_pair_group = expected_port_pair_groups.pop()
            else:
                expected_port_pair_group = self._get_expected_port_pair_group(
                    port_pair_group)
            with self.port_pair_group(port_pair_group=port_pair_group) as pg:
                for k, v in expected_port_pair_group.items():
                    self.assertEqual(pg['port_pair_group'][k], v)

    @staticmethod
    def _get_expected_port_chain(port_chain):
        chain_params = port_chain.get('chain_parameters') or dict()
        chain_params.setdefault('correlation', 'mpls')
        chain_params.setdefault('symmetric', False)
        ret = {
            'name': port_chain.get('name') or '',
            'description': port_chain.get('description') or '',
            'port_pair_groups': port_chain['port_pair_groups'],
            'flow_classifiers': port_chain.get('flow_classifiers') or [],
            'chain_parameters': chain_params
        }
        if port_chain.get('chain_id'):
            ret['chain_id'] = port_chain['chain_id']
        return ret

    def _test_create_port_chain(self, port_chain, expected_port_chain=None):
        if expected_port_chain is None:
            expected_port_chain = self._get_expected_port_chain(port_chain)
        with self.port_chain(port_chain=port_chain) as pc:
            for k, v in expected_port_chain.items():
                self.assertEqual(pc['port_chain'][k], v)

    def _test_create_port_chains(
        self, port_chains, expected_port_chains=None
    ):
        if port_chains:
            port_chain = port_chains.pop()
            if expected_port_chains:
                expected_port_chain = expected_port_chains.pop()
            else:
                expected_port_chain = self._get_expected_port_chain(
                    port_chain)
            with self.port_chain(port_chain=port_chain) as pc:
                for k, v in expected_port_chain.items():
                    self.assertEqual(pc['port_chain'][k], v)

    @staticmethod
    def _get_expected_graph(service_graph):
        ret = {
            'name': service_graph.get('name') or '',
            'description': service_graph.get('description') or '',
            'port_chains': service_graph.get('port_chains')
        }
        return ret

    def _test_create_service_graph(self, service_graph, expected_graph=None):
        if expected_graph is None:
            expected_graph = self._get_expected_graph(service_graph)
        with self.service_graph(service_graph=service_graph) as graph:
            for k, v in expected_graph.items():
                self.assertEqual(graph['service_graph'][k], v)


class SfcDbPluginTestCase(
    base.NeutronDbPluginV2TestCase,
    test_flowclassifier_db.FlowClassifierDbPluginTestCaseBase,
    SfcDbPluginTestCaseBase
):
    resource_prefix_map = dict([
        (k, sfc.SFC_PREFIX)
        for k in sfc.RESOURCE_ATTRIBUTE_MAP.keys()
    ] + [
        (k, fc_ext.FLOW_CLASSIFIER_PREFIX)
        for k in fc_ext.RESOURCE_ATTRIBUTE_MAP.keys()
    ] + [
        (k, sg_ext.SG_PREFIX)
        for k in sg_ext.RESOURCE_ATTRIBUTE_MAP.keys()
    ])

    def setUp(self, core_plugin=None, sfc_plugin=None,
              flowclassifier_plugin=None, ext_mgr=None):
        mock_log_p = mock.patch.object(sfc_db, 'LOG')
        self.mock_log = mock_log_p.start()
        cfg.CONF.register_opts(sfc.sfc_quota_opts, 'QUOTAS')
        if not sfc_plugin:
            sfc_plugin = DB_SFC_PLUGIN_CLASS
        if not flowclassifier_plugin:
            flowclassifier_plugin = (
                test_flowclassifier_db.DB_FLOWCLASSIFIER_PLUGIN_CLASS)

        service_plugins = {
            sfc.SFC_EXT: sfc_plugin,
            fc_ext.FLOW_CLASSIFIER_EXT: flowclassifier_plugin
        }
        sfc_db.SfcDbPlugin.supported_extension_aliases = [
            sfc.SFC_EXT, sg_ext.SG_EXT]
        sfc_db.SfcDbPlugin.path_prefix = sfc.SFC_PREFIX
        fdb.FlowClassifierDbPlugin.supported_extension_aliases = [
            fc_ext.FLOW_CLASSIFIER_EXT]
        fdb.FlowClassifierDbPlugin.path_prefix = (
            fc_ext.FLOW_CLASSIFIER_PREFIX
        )
        super(SfcDbPluginTestCase, self).setUp(
            ext_mgr=ext_mgr,
            plugin=core_plugin,
            service_plugins=service_plugins
        )
        if not ext_mgr:
            self.sfc_plugin = importutils.import_object(sfc_plugin)
            self.flowclassifier_plugin = importutils.import_object(
                flowclassifier_plugin)
            ext_mgr = api_ext.PluginAwareExtensionManager(
                extensions_path,
                {
                    sfc.SFC_EXT: self.sfc_plugin,
                    fc_ext.FLOW_CLASSIFIER_EXT: self.flowclassifier_plugin
                }
            )
            app = config.load_paste_app('extensions_test_app')
            self.ext_api = api_ext.ExtensionMiddleware(app, ext_mgr=ext_mgr)

    def test_create_port_chain(self):
        with self.port_pair_group(port_pair_group={}) as pg:
            self._test_create_port_chain({
                'port_pair_groups': [pg['port_pair_group']['id']]})

    def test_quota_create_port_chain(self):
        cfg.CONF.set_override('quota_port_chain', 3, group='QUOTAS')
        with self.port_pair_group(
            port_pair_group={}, do_delete=False
        ) as pg1, self.port_pair_group(
            port_pair_group={}, do_delete=False
        ) as pg2, self.port_pair_group(
            port_pair_group={}, do_delete=False
        ) as pg3, self.port_pair_group(
            port_pair_group={}, do_delete=False
        ) as pg4:
            self._create_port_chain(
                self.fmt, {
                    'port_pair_groups': [pg1['port_pair_group']['id']]
                }, expected_res_status=201)
            self._create_port_chain(
                self.fmt, {
                    'port_pair_groups': [pg2['port_pair_group']['id']]
                }, expected_res_status=201)
            self._create_port_chain(
                self.fmt, {
                    'port_pair_groups': [pg3['port_pair_group']['id']]
                }, expected_res_status=201)
            self._create_port_chain(
                self.fmt, {
                    'port_pair_groups': [pg4['port_pair_group']['id']]
                }, expected_res_status=409)

    def test_create_port_chain_all_fields(self):
        with self.port_pair_group(port_pair_group={}) as pg:
            self._test_create_port_chain({
                'port_pair_groups': [pg['port_pair_group']['id']],
                'flow_classifiers': [],
                'name': 'abc',
                'description': 'def',
                'chain_parameters': {'symmetric': False, 'correlation': 'mpls'}
            })

    def test_create_port_chain_all_fields_with_chain_id(self):
        with self.port_pair_group(port_pair_group={}) as pg:
            self._test_create_port_chain({
                'port_pair_groups': [pg['port_pair_group']['id']],
                'flow_classifiers': [],
                'name': 'abc',
                'description': 'def',
                'chain_parameters': {'symmetric': False,
                                     'correlation': 'mpls'},
                'chain_id': 99
            })

    def test_create_port_chain_all_fields_with_symmetric(self):
        with self.port_pair_group(port_pair_group={}) as pg:
            self._test_create_port_chain({
                'port_pair_groups': [pg['port_pair_group']['id']],
                'flow_classifiers': [],
                'name': 'abc',
                'description': 'def',
                'chain_parameters': {'symmetric': True, 'correlation': 'mpls'}
            })

    def test_create_port_chain_multi_port_pair_groups(self):
        with self.port_pair_group(
            port_pair_group={}
        ) as pg1, self.port_pair_group(
            port_pair_group={}
        ) as pg2:
            self._test_create_port_chain({
                'port_pair_groups': [
                    pg1['port_pair_group']['id'],
                    pg2['port_pair_group']['id']
                ]
            })

    def test_create_port_chain_shared_port_pair_groups(self):
        with self.port_pair_group(
            port_pair_group={}
        ) as pg1, self.port_pair_group(
            port_pair_group={}
        ) as pg2, self.port_pair_group(
            port_pair_group={}
        ) as pg3:
            self._test_create_port_chains([{
                'port_pair_groups': [
                    pg1['port_pair_group']['id'],
                    pg2['port_pair_group']['id']
                ]
            }, {
                'port_pair_groups': [
                    pg1['port_pair_group']['id'],
                    pg3['port_pair_group']['id']
                ]
            }])

    def test_create_port_chain_shared_port_pair_groups_different_order(self):
        with self.port_pair_group(
            port_pair_group={}
        ) as pg1, self.port_pair_group(
            port_pair_group={}
        ) as pg2:
            self._test_create_port_chains([{
                'port_pair_groups': [
                    pg1['port_pair_group']['id'],
                    pg2['port_pair_group']['id']
                ]
            }, {
                'port_pair_groups': [
                    pg2['port_pair_group']['id'],
                    pg1['port_pair_group']['id']
                ]
            }])

    def test_create_port_chain_with_empty_chain_parameters(self):
        with self.port_pair_group(port_pair_group={}) as pg:
            self._test_create_port_chain({
                'chain_parameters': {},
                'port_pair_groups': [pg['port_pair_group']['id']]
            })

    def test_create_port_chain_with_none_chain_parameters(self):
        with self.port_pair_group(port_pair_group={}) as pg:
            self._test_create_port_chain({
                'chain_parameters': None,
                'port_pair_groups': [pg['port_pair_group']['id']]
            })

    def test_create_port_chain_with_default_chain_parameters(self):
        with self.port_pair_group(port_pair_group={}) as pg:
            self._test_create_port_chain({
                'chain_parameters': {'symmetric': False,
                                     'correlation': 'mpls'},
                'port_pair_groups': [pg['port_pair_group']['id']]
            })

    def test_create_port_chains_with_conflicting_chain_ids(self):
        with self.port_pair_group(
            port_pair_group={}, do_delete=False
        ) as pg1, self.port_pair_group(
            port_pair_group={}, do_delete=False
        ) as pg2:
            self._create_port_chain(
                self.fmt, {
                    'port_pair_groups': [pg1['port_pair_group']['id']],
                    'chain_id': 88
                }, expected_res_status=201)
            self._create_port_chain(
                self.fmt, {
                    'port_pair_groups': [pg2['port_pair_group']['id']],
                    'chain_id': 88
                }, expected_res_status=400
            )

    def test_create_port_chain_with_none_flow_classifiers(self):
        with self.port_pair_group(port_pair_group={}) as pg:
            self._test_create_port_chain({
                'flow_classifiers': None,
                'port_pair_groups': [pg['port_pair_group']['id']]
            })

    def test_create_port_chain_with_empty_flow_classifiers(self):
        with self.port_pair_group(port_pair_group={}) as pg:
            self._test_create_port_chain({
                'flow_classifiers': [],
                'port_pair_groups': [pg['port_pair_group']['id']]
            })

    def test_create_port_chain_with_flow_classifiers(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'logical_source_port': port['port']['id']
            }) as fc:
                with self.port_pair_group(port_pair_group={}) as pg:
                    self._test_create_port_chain({
                        'flow_classifiers': [fc['flow_classifier']['id']],
                        'port_pair_groups': [pg['port_pair_group']['id']]
                    })

    def test_create_port_chain_with_multi_flow_classifiers(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'source_ip_prefix': '192.168.100.0/24',
                'logical_source_port': port['port']['id']
            }) as fc1, self.flow_classifier(flow_classifier={
                'source_ip_prefix': '192.168.101.0/24',
                'logical_source_port': port['port']['id']
            }) as fc2:
                with self.port_pair_group(port_pair_group={}) as pg:
                    self._test_create_port_chain({
                        'flow_classifiers': [
                            fc1['flow_classifier']['id'],
                            fc2['flow_classifier']['id']
                        ],
                        'port_pair_groups': [pg['port_pair_group']['id']]
                    })

    def test_create_port_chain_with_flow_classifiers_basic_the_same(self):
        with self.port(
            name='test1'
        ) as port1, self.port(
            name='test2'
        ) as port2:
            with self.flow_classifier(flow_classifier={
                'source_ip_prefix': '192.168.100.0/24',
                'logical_source_port': port1['port']['id']
            }) as fc1, self.flow_classifier(flow_classifier={
                'source_ip_prefix': '192.168.100.0/24',
                'logical_source_port': port2['port']['id']
            }) as fc2:
                with self.port_pair_group(port_pair_group={}) as pg:
                    self._test_create_port_chain({
                        'flow_classifiers': [
                            fc1['flow_classifier']['id'],
                            fc2['flow_classifier']['id']
                        ],
                        'port_pair_groups': [pg['port_pair_group']['id']]
                    })

    def test_create_multi_port_chain_with_flow_classifiers(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'source_ip_prefix': '192.168.100.0/24',
                'logical_source_port': port['port']['id']
            }) as fc1, self.flow_classifier(flow_classifier={
                'source_ip_prefix': '192.168.101.0/24',
                'logical_source_port': port['port']['id']
            }) as fc2:
                with self.port_pair_group(
                    port_pair_group={}
                ) as pg1, self.port_pair_group(
                    port_pair_group={}
                ) as pg2:
                    with self.port_chain(
                        port_chain={
                            'flow_classifiers': [
                                fc1['flow_classifier']['id']
                            ],
                            'port_pair_groups': [
                                pg1['port_pair_group']['id']
                            ]
                        }
                    ):
                        self._test_create_port_chain({
                            'flow_classifiers': [
                                fc2['flow_classifier']['id']
                            ],
                            'port_pair_groups': [pg2['port_pair_group']['id']]
                        })

    def test_create_multi_port_chain_with_conflict_flow_classifiers(self):
        with self.port(
            name='test1'
        ) as port1, self.port(
            name='test2'
        ) as port2:
            with self.flow_classifier(flow_classifier={
                'source_ip_prefix': '192.168.100.0/24',
                'logical_source_port': port1['port']['id']
            }) as fc1, self.flow_classifier(flow_classifier={
                'source_ip_prefix': '192.168.100.0/24',
                'logical_source_port': port2['port']['id']
            }) as fc2:
                with self.port_pair_group(
                    port_pair_group={}
                ) as pg1, self.port_pair_group(
                    port_pair_group={}
                ) as pg2:
                    with self.port_chain(
                        port_chain={
                            'flow_classifiers': [
                                fc1['flow_classifier']['id']
                            ],
                            'port_pair_groups': [
                                pg1['port_pair_group']['id']
                            ]
                        }
                    ):
                        self._create_port_chain(
                            self.fmt, {
                                'flow_classifiers': [
                                    fc2['flow_classifier']['id']
                                ],
                                'port_pair_groups': [
                                    pg2['port_pair_group']['id']
                                ]
                            },
                            expected_res_status=400
                        )

    def test_create_multi_port_chain_with_same_flow_classifier(self):
        with self.port(
            name='test1'
        ) as port1:
            with self.flow_classifier(flow_classifier={
                'source_ip_prefix': '192.168.100.0/24',
                'logical_source_port': port1['port']['id']
            }) as fc:
                with self.port_pair_group(
                    port_pair_group={}
                ) as pg1, self.port_pair_group(
                    port_pair_group={}
                ) as pg2:
                    with self.port_chain(
                        port_chain={
                            'flow_classifiers': [
                                fc['flow_classifier']['id']
                            ],
                            'port_pair_groups': [
                                pg1['port_pair_group']['id']
                            ]
                        }
                    ):
                        self._create_port_chain(
                            self.fmt, {
                                'flow_classifiers': [
                                    fc['flow_classifier']['id']
                                ],
                                'port_pair_groups': [
                                    pg2['port_pair_group']['id']
                                ]
                            },
                            expected_res_status=409
                        )

    def test_create_port_chain_with_port_pairs(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id']
            }) as pp1, self.port_pair(port_pair={
                'ingress': dst_port['port']['id'],
                'egress': src_port['port']['id']
            }) as pp2:
                with self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp1['port_pair']['id']
                    ]
                }) as pg1, self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp2['port_pair']['id']
                    ]
                }) as pg2:
                    self._test_create_port_chain({
                        'port_pair_groups': [
                            pg1['port_pair_group']['id'],
                            pg2['port_pair_group']['id']
                        ]
                    })

    def test_create_port_chain_with_empty_port_pair_groups(self):
        self._create_port_chain(
            self.fmt, {'port_pair_groups': []},
            expected_res_status=400
        )

    def test_create_port_chain_with_nonuuid_port_pair_group_id(self):
        self._create_port_chain(
            self.fmt, {'port_pair_groups': ['unknown']},
            expected_res_status=400
        )

    def test_create_port_chain_with_unknown_port_pair_group_id(self):
        self._create_port_chain(
            self.fmt, {'port_pair_groups': [uuidutils.generate_uuid()]},
            expected_res_status=404
        )

    def test_create_port_chain_with_same_port_pair_groups(self):
        with self.port_pair_group(
            port_pair_group={}
        ) as pg:
            with self.port_chain(
                port_chain={
                    'port_pair_groups': [pg['port_pair_group']['id']]
                }
            ):
                self._create_port_chain(
                    self.fmt, {
                        'port_pair_groups': [pg['port_pair_group']['id']]
                    }, expected_res_status=409
                )

    def test_create_port_chain_with_no_port_pair_groups(self):
        self._create_port_chain(
            self.fmt, {}, expected_res_status=400
        )

    def test_create_port_chain_with_invalid_chain_parameters(self):
        with self.port_pair_group(port_pair_group={}) as pg:
            self._create_port_chain(
                self.fmt, {
                    'chain_parameters': {'correlation': 'unknown'},
                    'port_pair_groups': [pg['port_pair_group']['id']]
                }, expected_res_status=400
            )

    def test_create_port_chain_with_invalid_chain_parameters_symmetric(self):
        with self.port_pair_group(port_pair_group={}) as pg:
            self._create_port_chain(
                self.fmt, {
                    'chain_parameters': {'symmetric': 'abc'},
                    'port_pair_groups': [pg['port_pair_group']['id']]
                }, expected_res_status=400
            )

    def test_create_port_chain_unknown_flow_classifiers(self):
        with self.port_pair_group(port_pair_group={}) as pg:
            self._create_port_chain(
                self.fmt, {
                    'flow_classifiers': [uuidutils.generate_uuid()],
                    'port_pair_groups': [pg['port_pair_group']['id']]
                }, expected_res_status=404
            )

    def test_create_port_chain_nouuid_flow_classifiers(self):
        with self.port_pair_group(port_pair_group={}) as pg:
            self._create_port_chain(
                self.fmt, {
                    'flow_classifiers': ['unknown'],
                    'port_pair_groups': [pg['port_pair_group']['id']]
                }, expected_res_status=400
            )

    def test_list_port_chains(self):
        with self.port_pair_group(
            port_pair_group={}
        ) as pg1, self.port_pair_group(
            port_pair_group={}
        ) as pg2:
            with self.port_chain(port_chain={
                'port_pair_groups': [pg1['port_pair_group']['id']]
            }) as pc1, self.port_chain(port_chain={
                'port_pair_groups': [pg2['port_pair_group']['id']]
            }) as pc2:
                port_chains = [pc1, pc2]
                self._test_list_resources(
                    'port_chain', port_chains
                )

    def test_list_port_chains_with_params(self):
        with self.port_pair_group(
            port_pair_group={}
        ) as pg1, self.port_pair_group(
            port_pair_group={}
        ) as pg2:
            with self.port_chain(port_chain={
                'name': 'test1',
                'port_pair_groups': [pg1['port_pair_group']['id']]
            }) as pc1, self.port_chain(port_chain={
                'name': 'test2',
                'port_pair_groups': [pg2['port_pair_group']['id']]
            }) as pc2:
                self._test_list_resources(
                    'port_chain', [pc1],
                    query_params='name=test1'
                )
                self._test_list_resources(
                    'port_chain', [pc2],
                    query_params='name=test2'
                )
                self._test_list_resources(
                    'port_chain', [],
                    query_params='name=test3'
                )

    def test_list_port_chains_with_unknown_params(self):
        with self.port_pair_group(
            port_pair_group={}
        ) as pg1, self.port_pair_group(
            port_pair_group={}
        ) as pg2:
            with self.port_chain(port_chain={
                'name': 'test1',
                'port_pair_groups': [pg1['port_pair_group']['id']]
            }) as pc1, self.port_chain(port_chain={
                'name': 'test2',
                'port_pair_groups': [pg2['port_pair_group']['id']]
            }) as pc2:
                self._test_list_resources(
                    'port_chain', [pc1, pc2],
                    query_params='hello=test3'
                )

    def test_show_port_chain(self):
        with self.port_pair_group(
            port_pair_group={}
        ) as pg:
            with self.port_chain(port_chain={
                'name': 'test1',
                'description': 'portchain',
                'port_pair_groups': [pg['port_pair_group']['id']]
            }) as pc:
                req = self.new_show_request(
                    'port_chains', pc['port_chain']['id']
                )
                res = self.deserialize(
                    self.fmt, req.get_response(self.ext_api)
                )
                expected = self._get_expected_port_chain(pc['port_chain'])
                self._assert_port_chain_equal(res['port_chain'], expected)

    def test_show_port_chain_noexist(self):
        req = self.new_show_request(
            'port_chains', '1'
        )
        res = req.get_response(self.ext_api)
        self.assertEqual(404, res.status_int)

    def test_update_port_chain_add_flow_classifiers(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(
                flow_classifier={
                    'source_ip_prefix': '192.168.100.0/24',
                    'logical_source_port': port['port']['id']
                }
            ) as fc1, self.flow_classifier(
                flow_classifier={
                    'source_ip_prefix': '192.168.101.0/24',
                    'logical_source_port': port['port']['id']
                }
            ) as fc2:
                with self.port_pair_group(
                    port_pair_group={}
                ) as pg:
                    with self.port_chain(port_chain={
                        'name': 'test1',
                        'description': 'desc1',
                        'port_pair_groups': [pg['port_pair_group']['id']],
                        'flow_classifiers': [fc1['flow_classifier']['id']]
                    }) as pc:
                        updates = {
                            'name': 'test2',
                            'description': 'desc2',
                            'flow_classifiers': [
                                fc1['flow_classifier']['id'],
                                fc2['flow_classifier']['id']
                            ]
                        }
                        req = self.new_update_request(
                            'port_chains', {'port_chain': updates},
                            pc['port_chain']['id']
                        )
                        res = self.deserialize(
                            self.fmt,
                            req.get_response(self.ext_api)
                        )
                        expected = pc['port_chain']
                        expected.update(updates)
                        self._assert_port_chain_equal(
                            res['port_chain'], expected
                        )
                        req = self.new_show_request(
                            'port_chains', pc['port_chain']['id']
                        )
                        res = self.deserialize(
                            self.fmt, req.get_response(self.ext_api)
                        )
                        self._assert_port_chain_equal(
                            res['port_chain'], expected
                        )

    def test_update_port_chain_remove_flow_classifiers(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(
                flow_classifier={
                    'source_ip_prefix': '192.168.100.0/24',
                    'logical_source_port': port['port']['id']
                }
            ) as fc1, self.flow_classifier(
                flow_classifier={
                    'source_ip_prefix': '192.168.101.0/24',
                    'logical_source_port': port['port']['id']
                }
            ) as fc2:
                with self.port_pair_group(
                    port_pair_group={}
                ) as pg:
                    with self.port_chain(port_chain={
                        'name': 'test1',
                        'description': 'desc1',
                        'port_pair_groups': [pg['port_pair_group']['id']],
                        'flow_classifiers': [
                            fc1['flow_classifier']['id'],
                            fc2['flow_classifier']['id']
                        ]
                    }) as pc:
                        updates = {
                            'name': 'test2',
                            'description': 'desc2',
                            'flow_classifiers': [
                                fc1['flow_classifier']['id']
                            ]
                        }
                        req = self.new_update_request(
                            'port_chains', {'port_chain': updates},
                            pc['port_chain']['id']
                        )
                        res = self.deserialize(
                            self.fmt,
                            req.get_response(self.ext_api)
                        )
                        expected = pc['port_chain']
                        expected.update(updates)
                        self._assert_port_chain_equal(
                            res['port_chain'], expected
                        )
                        req = self.new_show_request(
                            'port_chains', pc['port_chain']['id']
                        )
                        res = self.deserialize(
                            self.fmt, req.get_response(self.ext_api)
                        )
                        self._assert_port_chain_equal(
                            res['port_chain'], expected
                        )

    def test_update_port_chain_replace_flow_classifiers(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(
                flow_classifier={
                    'source_ip_prefix': '192.168.100.0/24',
                    'logical_source_port': port['port']['id']
                }
            ) as fc1, self.flow_classifier(
                flow_classifier={
                    'source_ip_prefix': '192.168.101.0/24',
                    'logical_source_port': port['port']['id']
                }
            ) as fc2:
                with self.port_pair_group(
                    port_pair_group={}
                ) as pg:
                    with self.port_chain(port_chain={
                        'name': 'test1',
                        'description': 'desc1',
                        'port_pair_groups': [pg['port_pair_group']['id']],
                        'flow_classifiers': [fc1['flow_classifier']['id']]
                    }) as pc:
                        updates = {
                            'name': 'test2',
                            'description': 'desc2',
                            'flow_classifiers': [fc2['flow_classifier']['id']]
                        }
                        req = self.new_update_request(
                            'port_chains', {'port_chain': updates},
                            pc['port_chain']['id']
                        )
                        res = self.deserialize(
                            self.fmt,
                            req.get_response(self.ext_api)
                        )
                        expected = pc['port_chain']
                        expected.update(updates)
                        self._assert_port_chain_equal(
                            res['port_chain'], expected
                        )
                        req = self.new_show_request(
                            'port_chains', pc['port_chain']['id']
                        )
                        res = self.deserialize(
                            self.fmt, req.get_response(self.ext_api)
                        )
                        self._assert_port_chain_equal(
                            res['port_chain'], expected
                        )

    def test_update_port_chain_flow_classifiers_basic_the_same(self):
        with self.port(
            name='test1'
        ) as port1, self.port(
            name='test2'
        ) as port2:
            with self.flow_classifier(
                flow_classifier={
                    'source_ip_prefix': '192.168.100.0/24',
                    'logical_source_port': port1['port']['id']
                }
            ) as fc1, self.flow_classifier(
                flow_classifier={
                    'source_ip_prefix': '192.168.100.0/24',
                    'logical_source_port': port2['port']['id']
                }
            ) as fc2:
                with self.port_pair_group(
                    port_pair_group={}
                ) as pg:
                    with self.port_chain(port_chain={
                        'name': 'test1',
                        'description': 'desc1',
                        'port_pair_groups': [pg['port_pair_group']['id']],
                        'flow_classifiers': [fc1['flow_classifier']['id']]
                    }) as pc:
                        updates = {
                            'name': 'test2',
                            'description': 'desc2',
                            'flow_classifiers': [fc2['flow_classifier']['id']]
                        }
                        req = self.new_update_request(
                            'port_chains', {'port_chain': updates},
                            pc['port_chain']['id']
                        )
                        res = self.deserialize(
                            self.fmt,
                            req.get_response(self.ext_api)
                        )
                        expected = pc['port_chain']
                        expected.update(updates)
                        self._assert_port_chain_equal(
                            res['port_chain'], expected
                        )
                        req = self.new_show_request(
                            'port_chains', pc['port_chain']['id']
                        )
                        res = self.deserialize(
                            self.fmt, req.get_response(self.ext_api)
                        )
                        self._assert_port_chain_equal(
                            res['port_chain'], expected
                        )

    def test_update_port_chain_conflict_flow_classifiers(self):
        with self.port(
            name='test1'
        ) as port1, self.port(
            name='test2'
        ) as port2:
            with self.flow_classifier(
                flow_classifier={
                    'source_ip_prefix': '192.168.100.0/24',
                    'logical_source_port': port1['port']['id']
                }
            ) as fc1, self.flow_classifier(
                flow_classifier={
                    'source_ip_prefix': '192.168.101.0/24',
                    'logical_source_port': port1['port']['id']
                }
            ) as fc2, self.flow_classifier(
                flow_classifier={
                    'source_ip_prefix': '192.168.100.0/24',
                    'logical_source_port': port2['port']['id']
                }
            ) as fc3:
                with self.port_pair_group(
                    port_pair_group={}
                ) as pg1, self.port_pair_group(
                    port_pair_group={}
                ) as pg2:
                    with self.port_chain(port_chain={
                        'port_pair_groups': [pg1['port_pair_group']['id']],
                        'flow_classifiers': [fc1['flow_classifier']['id']]
                    }), self.port_chain(port_chain={
                        'name': 'test2',
                        'port_pair_groups': [pg2['port_pair_group']['id']],
                        'flow_classifiers': [fc2['flow_classifier']['id']]
                    }) as pc2:
                        updates = {
                            'flow_classifiers': [fc3['flow_classifier']['id']]
                        }
                        req = self.new_update_request(
                            'port_chains', {'port_chain': updates},
                            pc2['port_chain']['id']
                        )
                        res = req.get_response(self.ext_api)
                        self.assertEqual(400, res.status_int)

    def test_update_port_chain_add_port_pair_groups(self):
        with self.port_pair_group(
            port_pair_group={}
        ) as pg1, self.port_pair_group(
            port_pair_group={}
        ) as pg2:
            with self.port_chain(port_chain={
                'port_pair_groups': [pg1['port_pair_group']['id']],
            }) as pc:
                updates = {
                    'port_pair_groups': [
                        pg1['port_pair_group']['id'],
                        pg2['port_pair_group']['id']
                    ]
                }
                req = self.new_update_request(
                    'port_chains', {'port_chain': updates},
                    pc['port_chain']['id']
                )
                res = self.deserialize(
                    self.fmt,
                    req.get_response(self.ext_api)
                )
                expected = pc['port_chain']
                expected.update(updates)
                self._assert_port_chain_equal(res['port_chain'], expected)
                req = self.new_show_request(
                    'port_chains', pc['port_chain']['id']
                )
                res = self.deserialize(
                    self.fmt, req.get_response(self.ext_api)
                )
                self._assert_port_chain_equal(res['port_chain'], expected)

    def test_update_port_chain_remove_port_pair_groups(self):
        with self.port_pair_group(
            port_pair_group={}
        ) as pg1, self.port_pair_group(
            port_pair_group={}
        ) as pg2:
            with self.port_chain(port_chain={
                'port_pair_groups': [
                    pg1['port_pair_group']['id'],
                    pg2['port_pair_group']['id'],
                ],
            }) as pc:
                updates = {
                    'port_pair_groups': [
                        pg1['port_pair_group']['id']
                    ]
                }
                req = self.new_update_request(
                    'port_chains', {'port_chain': updates},
                    pc['port_chain']['id']
                )
                res = self.deserialize(
                    self.fmt,
                    req.get_response(self.ext_api)
                )
                expected = pc['port_chain']
                expected.update(updates)
                self._assert_port_chain_equal(res['port_chain'], expected)
                req = self.new_show_request(
                    'port_chains', pc['port_chain']['id']
                )
                res = self.deserialize(
                    self.fmt, req.get_response(self.ext_api)
                )
                self._assert_port_chain_equal(res['port_chain'], expected)

    def test_update_port_chain_replace_port_pair_groups(self):
        with self.port_pair_group(
            port_pair_group={}
        ) as pg1, self.port_pair_group(
            port_pair_group={}
        ) as pg2:
            with self.port_chain(port_chain={
                'port_pair_groups': [pg1['port_pair_group']['id']],
            }) as pc:
                updates = {
                    'port_pair_groups': [pg2['port_pair_group']['id']]
                }
                req = self.new_update_request(
                    'port_chains', {'port_chain': updates},
                    pc['port_chain']['id']
                )
                res = self.deserialize(
                    self.fmt,
                    req.get_response(self.ext_api)
                )
                expected = pc['port_chain']
                expected.update(updates)
                self._assert_port_chain_equal(res['port_chain'], expected)
                req = self.new_show_request(
                    'port_chains', pc['port_chain']['id']
                )
                res = self.deserialize(
                    self.fmt, req.get_response(self.ext_api)
                )
                self._assert_port_chain_equal(res['port_chain'], expected)

    def test_update_port_chain_chain_parameters(self):
        with self.port_pair_group(
            port_pair_group={}
        ) as pg:
            with self.port_chain(port_chain={
                'port_pair_groups': [pg['port_pair_group']['id']],
            }) as pc:
                updates = {
                    'chain_parameters': {'correlation': 'mpls'}
                }
                req = self.new_update_request(
                    'port_chains', {'port_chain': updates},
                    pc['port_chain']['id']
                )
                res = req.get_response(self.ext_api)
                self.assertEqual(400, res.status_int)

    def test_update_port_chain_part_of_graph_fail(self):
        with self.port_pair_group(
            port_pair_group={}
        ) as pg1, self.port_pair_group(
            port_pair_group={}
        ) as pg2:
            with self.port_chain(port_chain={
                'port_pair_groups': [pg1['port_pair_group']['id']]
            }) as pc1, self.port_chain(port_chain={
                'port_pair_groups': [pg2['port_pair_group']['id']]
            }) as pc2:
                with self.service_graph(service_graph={
                    'name': 'test1',
                    'port_chains': {
                        pc1['port_chain']['id']: [pc2['port_chain']['id']]}
                }):
                    updates = {
                        'port_pair_groups': [uuidutils.generate_uuid()]
                    }
                    req = self.new_update_request(
                        'port_chains', {'port_chain': updates},
                        pc1['port_chain']['id']
                    )
                    res = req.get_response(self.ext_api)
                    self.assertEqual(409, res.status_int)
                    updates = {
                        'flow_classifiers': [uuidutils.generate_uuid()]
                    }
                    req = self.new_update_request(
                        'port_chains', {'port_chain': updates},
                        pc2['port_chain']['id']
                    )
                    res = req.get_response(self.ext_api)
                    self.assertEqual(409, res.status_int)
                    updates = {
                        'name': 'new name',
                        'description': 'new description'
                    }
                    req = self.new_update_request(
                        'port_chains', {'port_chain': updates},
                        pc1['port_chain']['id']
                    )
                    res = req.get_response(self.ext_api)
                    self.assertEqual(200, res.status_int)

    def test_delete_port_chain(self):
        with self.port_pair_group(
            port_pair_group={}
        ) as pg:
            with self.port_chain(port_chain={
                'port_pair_groups': [pg['port_pair_group']['id']]
            }, do_delete=False) as pc:
                req = self.new_delete_request(
                    'port_chains', pc['port_chain']['id']
                )
                res = req.get_response(self.ext_api)
                self.assertEqual(204, res.status_int)
                req = self.new_show_request(
                    'port_chains', pc['port_chain']['id']
                )
                res = req.get_response(self.ext_api)
                self.assertEqual(404, res.status_int)
                req = self.new_show_request(
                    'port_pair_groups', pg['port_pair_group']['id']
                )
                res = req.get_response(self.ext_api)
                self.assertEqual(200, res.status_int)

    def test_delete_port_chain_noexist(self):
        req = self.new_delete_request(
            'port_chains', '1'
        )
        res = req.get_response(self.ext_api)
        self.assertEqual(404, res.status_int)

    def test_delete_port_chain_part_of_graph_fail(self):
        with self.port_pair_group(
            port_pair_group={}
        ) as pg1, self.port_pair_group(
            port_pair_group={}
        ) as pg2:
            with self.port_chain(port_chain={
                'port_pair_groups': [pg1['port_pair_group']['id']]
            }) as pc1, self.port_chain(port_chain={
                'port_pair_groups': [pg2['port_pair_group']['id']]
            }) as pc2:
                with self.service_graph(service_graph={
                    'name': 'test1',
                    'port_chains': {
                        pc1['port_chain']['id']: [pc2['port_chain']['id']]}
                }):
                    req = self.new_delete_request(
                        'port_chains', pc1['port_chain']['id']
                    )
                    res = req.get_response(self.ext_api)
                    self.assertEqual(409, res.status_int)
                    req = self.new_delete_request(
                        'port_chains', pc2['port_chain']['id']
                    )
                    res = req.get_response(self.ext_api)
                    self.assertEqual(409, res.status_int)

    def test_delete_flow_classifier_port_chain_exist(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'logical_source_port': port['port']['id']
            }) as fc:
                with self.port_pair_group(port_pair_group={
                }) as pg:
                    with self.port_chain(port_chain={
                        'port_pair_groups': [pg['port_pair_group']['id']],
                        'flow_classifiers': [fc['flow_classifier']['id']]
                    }):
                        req = self.new_delete_request(
                            'flow_classifiers', fc['flow_classifier']['id']
                        )
                        res = req.get_response(self.ext_api)
                        self.assertEqual(409, res.status_int)

    def test_create_port_pair_group(self):
        self._test_create_port_pair_group({})

    def test_quota_create_port_pair_group_quota(self):
        cfg.CONF.set_override('quota_port_pair_group', 3, group='QUOTAS')
        self._create_port_pair_group(
            self.fmt, {'port_pairs': []}, expected_res_status=201
        )
        self._create_port_pair_group(
            self.fmt, {'port_pairs': []}, expected_res_status=201
        )
        self._create_port_pair_group(
            self.fmt, {'port_pairs': []}, expected_res_status=201
        )
        self._create_port_pair_group(
            self.fmt, {'port_pairs': []}, expected_res_status=409
        )

    def test_create_port_pair_group_all_fields(self):
        self._test_create_port_pair_group({
            'name': 'test1',
            'description': 'desc1',
            'port_pairs': [],
            'port_pair_group_parameters': {
                'lb_fields': ['ip_src', 'ip_dst'],
                'ppg_n_tuple_mapping': {
                    'ingress_n_tuple': {'source_ip_prefix': None},
                    'egress_n_tuple': {'destination_ip_prefix': None}}
            }
        })

    def test_create_port_pair_group_with_empty_parameters(self):
        self._test_create_port_pair_group({
            'name': 'test1',
            'description': 'desc1',
            'port_pairs': [],
            'port_pair_group_parameters': {}
        })

    def test_create_port_pair_group_with_none_parameters(self):
        self._test_create_port_pair_group({
            'name': 'test1',
            'description': 'desc1',
            'port_pairs': [],
            'port_pair_group_parameters': None
        })

    def test_create_port_pair_group_with_default_parameters(self):
        self._test_create_port_pair_group({
            'name': 'test1',
            'description': 'desc1',
            'port_pairs': [],
            'port_pair_group_parameters': {
                'lb_fields': [],
                'ppg_n_tuple_mapping': {}
            }
        })

    def test_create_port_pair_group_with_port_pairs(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id']
            }) as pp1, self.port_pair(port_pair={
                'ingress': dst_port['port']['id'],
                'egress': src_port['port']['id']
            }) as pp2:
                self._test_create_port_pair_group({
                    'port_pairs': [
                        pp1['port_pair']['id'],
                        pp2['port_pair']['id']
                    ]
                })

    def test_create_port_pair_group_consistent_correlations(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id'],
                'service_function_parameters': {'correlation': 'mpls'}
            }) as pp1, self.port_pair(port_pair={
                'ingress': dst_port['port']['id'],
                'egress': src_port['port']['id'],
                'service_function_parameters': {'correlation': 'mpls'}
            }) as pp2:
                self._test_create_port_pair_group({
                    'port_pairs': [
                        pp1['port_pair']['id'],
                        pp2['port_pair']['id']
                    ]
                })

    def test_create_port_pair_group_inconsistent_correlations(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id'],
                'service_function_parameters': {'correlation': 'mpls'}
            }) as pp1, self.port_pair(port_pair={
                'ingress': dst_port['port']['id'],
                'egress': src_port['port']['id'],
                'service_function_parameters': {'correlation': None}
            }) as pp2:
                self._create_port_pair_group(
                    self.fmt,
                    {'port_pairs': [
                        pp1['port_pair']['id'],
                        pp2['port_pair']['id']
                    ]},
                    expected_res_status=400)

    def test_create_port_pair_group_with_nouuid_port_pair_id(self):
        self._create_port_pair_group(
            self.fmt, {'port_pairs': ['unknown']},
            expected_res_status=400
        )

    def test_create_port_pair_group_with_unknown_port_pair_id(self):
        self._create_port_pair_group(
            self.fmt, {'port_pairs': [uuidutils.generate_uuid()]},
            expected_res_status=404
        )

    def test_create_port_pair_group_share_port_pair_id(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id']
            }) as pp:
                with self.port_pair_group(port_pair_group={
                    'port_pairs': [pp['port_pair']['id']]
                }):
                    self._create_port_pair_group(
                        self.fmt, {'port_pairs': [pp['port_pair']['id']]},
                        expected_res_status=409
                    )

    def test_list_port_pair_groups(self):
        with self.port_pair_group(port_pair_group={
            'name': 'test1'
        }) as pc1, self.port_pair_group(port_pair_group={
            'name': 'test2'
        }) as pc2:
            port_pair_groups = [pc1, pc2]
            self._test_list_resources(
                'port_pair_group', port_pair_groups
            )

    def test_list_port_pair_groups_with_params(self):
        with self.port_pair_group(port_pair_group={
            'name': 'test1'
        }) as pc1, self.port_pair_group(port_pair_group={
            'name': 'test2'
        }) as pc2:
            self._test_list_resources(
                'port_pair_group', [pc1],
                query_params='name=test1'
            )
            self._test_list_resources(
                'port_pair_group', [pc2],
                query_params='name=test2'
            )
            self._test_list_resources(
                'port_pair_group', [],
                query_params='name=test3'
            )

    def test_list_port_pair_groups_with_unknown_params(self):
        with self.port_pair_group(port_pair_group={
            'name': 'test1'
        }) as pc1, self.port_pair_group(port_pair_group={
            'name': 'test2'
        }) as pc2:
            self._test_list_resources(
                'port_pair_group', [pc1, pc2],
                query_params='hello=test3'
            )

    def test_show_port_pair_group(self):
        with self.port_pair_group(port_pair_group={
            'name': 'test1'
        }) as pc:
            req = self.new_show_request(
                'port_pair_groups', pc['port_pair_group']['id']
            )
            res = self.deserialize(
                self.fmt, req.get_response(self.ext_api)
            )
            for k, v in pc['port_pair_group'].items():
                self.assertEqual(res['port_pair_group'][k], v)

    def test_show_port_pair_group_noexist(self):
        req = self.new_show_request(
            'port_pair_groups', '1'
        )
        res = req.get_response(self.ext_api)
        self.assertEqual(404, res.status_int)

    def test_update_port_pair_group(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id']
            }) as pp1, self.port_pair(port_pair={
                'ingress': dst_port['port']['id'],
                'egress': src_port['port']['id']
            }) as pp2:
                with self.port_pair_group(port_pair_group={
                    'name': 'test1',
                    'description': 'desc1',
                    'port_pairs': [pp1['port_pair']['id']]
                }) as pg:
                    updates = {
                        'name': 'test2',
                        'description': 'desc2',
                        'port_pairs': [pp2['port_pair']['id']]
                    }
                    req = self.new_update_request(
                        'port_pair_groups', {'port_pair_group': updates},
                        pg['port_pair_group']['id']
                    )
                    res = self.deserialize(
                        self.fmt,
                        req.get_response(self.ext_api)
                    )
                    expected = pg['port_pair_group']
                    expected.update(updates)
                    for k, v in expected.items():
                        self.assertEqual(res['port_pair_group'][k], v)
                    req = self.new_show_request(
                        'port_pair_groups', pg['port_pair_group']['id']
                    )
                    res = self.deserialize(
                        self.fmt, req.get_response(self.ext_api)
                    )
                    for k, v in expected.items():
                        self.assertEqual(res['port_pair_group'][k], v)

    def test_update_port_pair_group_consistency_checks(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as port1, self.port(
            name='port2',
            device_id='default'
        ) as port2, self.port(
            name='port3',
            device_id='default'
        ) as port3, self.port(
            name='port4',
            device_id='default'
        ) as port4:
            with self.port_pair(port_pair={
                'ingress': port1['port']['id'],
                'egress': port2['port']['id'],
                'service_function_parameters': {'correlation': 'mpls'}
            }) as pp1, self.port_pair(port_pair={
                'ingress': port2['port']['id'],
                'egress': port3['port']['id'],
                'service_function_parameters': {'correlation': 'mpls'}
            }) as pp2, self.port_pair(port_pair={
                'ingress': port3['port']['id'],
                'egress': port4['port']['id'],
                'service_function_parameters': {'correlation': None}
            }) as pp3, self.port_pair(port_pair={
                'ingress': port4['port']['id'],
                'egress': port1['port']['id'],
                'service_function_parameters': {'correlation': 'mpls'}
            }) as pp4:
                with self.port_pair_group(port_pair_group={
                    'name': 'test1',
                    'description': 'desc1',
                    'port_pairs': [pp1['port_pair']['id'],
                                   pp2['port_pair']['id']]
                }) as pg:
                    updates = {
                        'name': 'test2',
                        'description': 'desc2',
                        'port_pairs': [pp1['port_pair']['id'],
                                       pp2['port_pair']['id'],
                                       pp3['port_pair']['id']]
                    }
                    req = self.new_update_request(
                        'port_pair_groups', {'port_pair_group': updates},
                        pg['port_pair_group']['id']
                    )
                    resp = req.get_response(self.ext_api)
                    self.assertEqual(400, resp.status_int)

                    updates = {
                        'name': 'test3',
                        'description': 'desc3',
                        'port_pairs': [pp1['port_pair']['id'],
                                       pp2['port_pair']['id'],
                                       pp4['port_pair']['id']]
                    }
                    req = self.new_update_request(
                        'port_pair_groups', {'port_pair_group': updates},
                        pg['port_pair_group']['id']
                    )
                    resp = req.get_response(self.ext_api)
                    res = self.deserialize(self.fmt, resp)
                    expected = pg['port_pair_group']
                    expected.update(updates)
                    for k, v in expected.items():
                        self.assertEqual(res['port_pair_group'][k], v)
                    req = self.new_show_request(
                        'port_pair_groups', pg['port_pair_group']['id']
                    )
                    res = self.deserialize(
                        self.fmt, req.get_response(self.ext_api)
                    )
                    for k, v in expected.items():
                        self.assertEqual(res['port_pair_group'][k], v)

    def test_delete_port_pair_group(self):
        with self.port_pair_group(port_pair_group={
            'name': 'test1'
        }, do_delete=False) as pc:
            req = self.new_delete_request(
                'port_pair_groups', pc['port_pair_group']['id']
            )
            res = req.get_response(self.ext_api)
            self.assertEqual(204, res.status_int)
            req = self.new_show_request(
                'port_pair_groups', pc['port_pair_group']['id']
            )
            res = req.get_response(self.ext_api)
            self.assertEqual(404, res.status_int)

    def test_delete_port_pair_group_port_chain_exist(self):
        with self.port_pair_group(port_pair_group={
            'name': 'test1'
        }) as pg:
            with self.port_chain(port_chain={
                'port_pair_groups': [pg['port_pair_group']['id']]
            }):
                req = self.new_delete_request(
                    'port_pair_groups', pg['port_pair_group']['id']
                )
                res = req.get_response(self.ext_api)
                self.assertEqual(409, res.status_int)

    def test_delete_port_pair_group_noexist(self):
        req = self.new_delete_request(
            'port_pair_groups', '1'
        )
        res = req.get_response(self.ext_api)
        self.assertEqual(404, res.status_int)

    def test_create_port_pair(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            self._test_create_port_pair({
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id']
            })

    def test_quota_create_port_pair_quota(self):
        cfg.CONF.set_override('quota_port_pair', 3, group='QUOTAS')
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port1, self.port(
            name='port2',
            device_id='default'
        ) as dst_port1, self.port(
            name='port3',
            device_id='default'
        ) as src_port2, self.port(
            name='port4',
            device_id='default'
        ) as dst_port2, self.port(
            name='port5',
            device_id='default'
        ) as src_port3, self.port(
            name='port6',
            device_id='default'
        ) as dst_port3, self.port(
            name='port7',
            device_id='default'
        ) as src_port4, self.port(
            name='port8',
            device_id='default'
        ) as dst_port4:
            self._create_port_pair(
                self.fmt, {
                    'ingress': src_port1['port']['id'],
                    'egress': dst_port1['port']['id']
                }, expected_res_status=201)
            self._create_port_pair(
                self.fmt, {
                    'ingress': src_port2['port']['id'],
                    'egress': dst_port2['port']['id']
                }, expected_res_status=201)
            self._create_port_pair(
                self.fmt, {
                    'ingress': src_port3['port']['id'],
                    'egress': dst_port3['port']['id']
                }, expected_res_status=201)
            self._create_port_pair(
                self.fmt, {
                    'ingress': src_port4['port']['id'],
                    'egress': dst_port4['port']['id']
                }, expected_res_status=409)

    def test_create_port_pair_all_fields(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            self._test_create_port_pair({
                'name': 'test1',
                'description': 'desc1',
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id'],
                'service_function_parameters': {
                    'correlation': None, 'weight': 2}
            })

    def test_create_port_pair_none_service_function_parameters(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            self._test_create_port_pair({
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id'],
                'service_function_parameters': None
            })

    def test_create_port_pair_empty_service_function_parameters(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            self._test_create_port_pair({
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id'],
                'service_function_parameters': {}
            })

    def test_create_port_pair_with_src_dst_same_port(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_dst_port:
            self._test_create_port_pair({
                'ingress': src_dst_port['port']['id'],
                'egress': src_dst_port['port']['id']
            })

    def test_create_port_pair_empty_input(self):
        self._create_port_pair(self.fmt, {}, expected_res_status=400)

    def test_create_port_pair_with_no_ingress(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as dst_port:
            self._create_port_pair(
                self.fmt,
                {
                    'egress': dst_port['port']['id']
                },
                expected_res_status=400
            )

    def test_create_port_pair_with_no_egress(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port:
            self._create_port_pair(
                self.fmt,
                {
                    'ingress': src_port['port']['id']
                },
                expected_res_status=400
            )

    def test_create_port_pair_with_nouuid_ingress(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as dst_port:
            self._create_port_pair(
                self.fmt,
                {
                    'ingress': '1',
                    'egress': dst_port['port']['id']
                },
                expected_res_status=400
            )

    def test_create_port_pair_with_unknown_ingress(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as dst_port:
            self._create_port_pair(
                self.fmt,
                {
                    'ingress': uuidutils.generate_uuid(),
                    'egress': dst_port['port']['id']
                },
                expected_res_status=404
            )

    def test_create_port_pair_with_nouuid_egress(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port:
            self._create_port_pair(
                self.fmt,
                {
                    'ingress': src_port['port']['id'],
                    'egress': '1'
                },
                expected_res_status=400
            )

    def test_create_port_pair_with_unknown_egress(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port:
            self._create_port_pair(
                self.fmt,
                {
                    'ingress': src_port['port']['id'],
                    'egress': uuidutils.generate_uuid()
                },
                expected_res_status=404
            )

    def test_create_port_pair_ingress_egress_different_hosts(self):
        with self.port(
            name='port1',
            device_id='device1'
        ) as src_port, self.port(
            name='port2',
            device_id='device2'
        ) as dst_port:
            self._create_port_pair(
                self.fmt,
                {
                    'ingress': src_port['port']['id'],
                    'egress': dst_port['port']['id']
                },
                expected_res_status=400
            )

    def test_create_port_pair_with_invalid_service_function_parameters(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_dst_port:
            self._create_port_pair(
                self.fmt,
                {
                    'ingress': src_dst_port['port']['id'],
                    'egress': src_dst_port['port']['id'],
                    'service_function_parameters': {'abc': 'def'}
                },
                expected_res_status=400
            )

    def test_create_port_pair_with_invalid_correlation(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_dst_port:
            self._create_port_pair(
                self.fmt,
                {
                    'ingress': src_dst_port['port']['id'],
                    'egress': src_dst_port['port']['id'],
                    'service_function_parameters': {'correlation': 'def'}
                },
                expected_res_status=400
            )

    def test_create_port_pair_with_invalid_weight(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_dst_port:
            self._create_port_pair(
                self.fmt,
                {
                    'ingress': src_dst_port['port']['id'],
                    'egress': src_dst_port['port']['id'],
                    'service_function_parameters': {'weight': -1}
                },
                expected_res_status=400
            )

    def test_list_port_pairs(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id']
            }) as pc1, self.port_pair(port_pair={
                'ingress': dst_port['port']['id'],
                'egress': src_port['port']['id']
            }) as pc2:
                port_pairs = [pc1, pc2]
                self._test_list_resources(
                    'port_pair', port_pairs
                )

    def test_list_port_pairs_with_params(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'name': 'test1',
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id']
            }) as pc1, self.port_pair(port_pair={
                'name': 'test2',
                'ingress': dst_port['port']['id'],
                'egress': src_port['port']['id']
            }) as pc2:
                self._test_list_resources(
                    'port_pair', [pc1],
                    query_params='name=test1'
                )
                self._test_list_resources(
                    'port_pair', [pc2],
                    query_params='name=test2'
                )
                self._test_list_resources(
                    'port_pair', [],
                    query_params='name=test3'
                )

    def test_list_port_pairs_with_unknown_params(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'name': 'test1',
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id']
            }) as pc1, self.port_pair(port_pair={
                'name': 'test2',
                'ingress': dst_port['port']['id'],
                'egress': src_port['port']['id']
            }) as pc2:
                port_pairs = [pc1, pc2]
                self._test_list_resources(
                    'port_pair', port_pairs,
                    query_params='hello=test3'
                )

    def test_show_port_pair(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id']
            }) as pc:
                req = self.new_show_request(
                    'port_pairs', pc['port_pair']['id']
                )
                res = self.deserialize(
                    self.fmt, req.get_response(self.ext_api)
                )
                for k, v in pc['port_pair'].items():
                    self.assertEqual(res['port_pair'][k], v)

    def test_show_port_pair_noexist(self):
        req = self.new_show_request(
            'port_pairs', '1'
        )
        res = req.get_response(self.ext_api)
        self.assertEqual(404, res.status_int)

    def test_update_port_pair(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'name': 'test1',
                'description': 'desc1',
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id']
            }) as pc:
                updates = {
                    'name': 'test2',
                    'description': 'desc2'
                }
                req = self.new_update_request(
                    'port_pairs', {'port_pair': updates},
                    pc['port_pair']['id']
                )
                res = self.deserialize(
                    self.fmt,
                    req.get_response(self.ext_api)
                )
                expected = pc['port_pair']
                expected.update(updates)
                for k, v in expected.items():
                    self.assertEqual(res['port_pair'][k], v)
                req = self.new_show_request(
                    'port_pairs', pc['port_pair']['id']
                )
                res = self.deserialize(
                    self.fmt, req.get_response(self.ext_api)
                )
                for k, v in expected.items():
                    self.assertEqual(res['port_pair'][k], v)

    def test_update_port_pair_service_function_parameters(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'name': 'test1',
                'description': 'desc1',
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id']
            }) as pc:
                updates = {
                    'service_function_parameters': {
                        'correlation': None, 'weight': 2,
                    }
                }
                req = self.new_update_request(
                    'port_pairs', {'port_pair': updates},
                    pc['port_pair']['id']
                )
                res = req.get_response(self.ext_api)
                self.assertEqual(400, res.status_int)

    def test_update_port_pair_ingress(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'name': 'test1',
                'description': 'desc1',
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id']
            }) as pc:
                updates = {
                    'ingress': dst_port['port']['id']
                }
                req = self.new_update_request(
                    'port_pairs', {'port_pair': updates},
                    pc['port_pair']['id']
                )
                res = req.get_response(self.ext_api)
                self.assertEqual(400, res.status_int)

    def test_update_port_pair_egress(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'name': 'test1',
                'description': 'desc1',
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id']
            }) as pc:
                updates = {
                    'egress': src_port['port']['id']
                }
                req = self.new_update_request(
                    'port_pairs', {'port_pair': updates},
                    pc['port_pair']['id']
                )
                res = req.get_response(self.ext_api)
                self.assertEqual(400, res.status_int)

    def test_delete_port_pair(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id']
            }, do_delete=False) as pc:
                req = self.new_delete_request(
                    'port_pairs', pc['port_pair']['id']
                )
                res = req.get_response(self.ext_api)
                self.assertEqual(204, res.status_int)
                req = self.new_show_request(
                    'port_pairs', pc['port_pair']['id']
                )
                res = req.get_response(self.ext_api)
                self.assertEqual(404, res.status_int)

    def test_delete_port_pair_noexist(self):
        req = self.new_delete_request(
            'port_pairs', '1'
        )
        res = req.get_response(self.ext_api)
        self.assertEqual(404, res.status_int)

    def test_delete_port_pair_port_pair_group_exist(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id']
            }) as pp:
                with self.port_pair_group(port_pair_group={
                    'port_pairs': [pp['port_pair']['id']]
                }):
                    req = self.new_delete_request(
                        'port_pairs', pp['port_pair']['id']
                    )
                    res = req.get_response(self.ext_api)
                    self.assertEqual(409, res.status_int)

    def test_delete_ingress_port_pair_exist(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id']
            }):
                req = self.new_delete_request(
                    'ports', src_port['port']['id']
                )
                res = req.get_response(self.api)
                self.assertEqual(500, res.status_int)

    def test_delete_egress_port_pair_exist(self):
        with self.port(
            name='port1',
            device_id='default'
        ) as src_port, self.port(
            name='port2',
            device_id='default'
        ) as dst_port:
            with self.port_pair(port_pair={
                'ingress': src_port['port']['id'],
                'egress': dst_port['port']['id']
            }):
                req = self.new_delete_request(
                    'ports', dst_port['port']['id']
                )
                res = req.get_response(self.api)
                self.assertEqual(500, res.status_int)

    def _test_create_service_graph_branching_ppg(
            self, src_corr, dst_corr, status):
        with self.port(
            name='port1',
            device_id='default'
        ) as port1, self.port(
            name='port2',
            device_id='default'
        ) as port2, self.port(
            name='port3',
            device_id='default'
        ) as port3, self.port(
            name='port4',
            device_id='default'
        ) as port4:
            with self.port_pair(port_pair={
                'ingress': port1['port']['id'],
                'egress': port1['port']['id'],
                'service_function_parameters': {'correlation': 'mpls'}
            }, do_delete=False) as pp1, self.port_pair(port_pair={
                'ingress': port2['port']['id'],
                'egress': port2['port']['id'],
                'service_function_parameters': {'correlation': src_corr}
            }, do_delete=False) as pp2, self.port_pair(port_pair={
                'ingress': port3['port']['id'],
                'egress': port3['port']['id'],
                'service_function_parameters': {'correlation': dst_corr}
            }, do_delete=False) as pp3, self.port_pair(port_pair={
                'ingress': port4['port']['id'],
                'egress': port4['port']['id'],
                'service_function_parameters': {'correlation': 'mpls'}
            }, do_delete=False) as pp4:
                with self.port_pair_group(
                    port_pair_group={'port_pairs': [pp1['port_pair']['id']]},
                    do_delete=False
                ) as pg1, self.port_pair_group(
                    port_pair_group={'port_pairs': [pp2['port_pair']['id']]},
                    do_delete=False
                ) as pg2, self.port_pair_group(
                    port_pair_group={'port_pairs': [pp3['port_pair']['id']]},
                    do_delete=False
                ) as pg3, self.port_pair_group(
                    port_pair_group={'port_pairs': [pp4['port_pair']['id']]},
                    do_delete=False
                ) as pg4:
                    with self.port_chain(
                        do_delete=False,
                        port_chain={'port_pair_groups': [
                            pg1['port_pair_group']['id'],
                            pg2['port_pair_group']['id']]}
                    ) as pc1, self.port_chain(
                        do_delete=False,
                        port_chain={'port_pair_groups': [
                            pg3['port_pair_group']['id'],
                            pg4['port_pair_group']['id']]}
                    ) as pc2:
                        self._create_service_graph(self.fmt, {
                            'port_chains': {
                                pc1['port_chain']['id']: [
                                    pc2['port_chain']['id']]
                            },
                            'name': 'abc',
                            'description': 'def'
                        }, expected_res_status=status)

    def test_create_service_graph_branching_ppg_no_src_corr_fail(self):
        self._test_create_service_graph_branching_ppg(None, 'mpls', 400)

    def test_create_service_graph_branching_ppg_no_dst_corr_fail(self):
        self._test_create_service_graph_branching_ppg('mpls', None, 400)

    def test_create_service_graph_branching_ppg_both_corrs_ok(self):
        self._test_create_service_graph_branching_ppg('mpls', 'mpls', 201)

    def test_create_service_graph_linear_dependency_only(self):
        # this test will create a graph consisting of 1 port chain being
        # dependent on 1 other port chain, thus with no branching.
        with self.port_pair_group(
            port_pair_group={}, do_delete=False
        ) as pg1, self.port_pair_group(
            port_pair_group={}, do_delete=False
        ) as pg2:
            with self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg1['port_pair_group']['id']]}
            ) as pc1, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg2['port_pair_group']['id']]}
            ) as pc2:
                self._create_service_graph(self.fmt, {
                    'port_chains': {
                        pc1['port_chain']['id']: [pc2['port_chain']['id']]
                    },
                    'name': 'abc',
                    'description': 'def'
                }, expected_res_status=201)

    def test_create_service_graph_branching_no_class(self):
        # this test will create a graph where 1 port chain will act
        # as a dependency to 2 other port chains, effectively
        # creating a branching service function chain.
        with self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg1, self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg2, self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg3:
            with self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg1['port_pair_group']['id']]}
            ) as pc1, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg2['port_pair_group']['id']]}
            ) as pc2, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg3['port_pair_group']['id']]}
            ) as pc3:
                self._create_service_graph(self.fmt, {
                    'port_chains': {
                        pc1['port_chain']['id']: [
                            pc2['port_chain']['id'],
                            pc3['port_chain']['id']
                        ]
                    },
                    'name': 'abc',
                    'description': 'def'
                }, expected_res_status=201)

    def test_create_service_graph_same_chain_fail(self):
        # this test will attempt to create a graph with a single branching
        # point having 2 port chains - which are actually the same port chain.
        with self.port_pair_group(
            port_pair_group={}
        ) as pg1, self.port_pair_group(
            port_pair_group={}
        ) as pg2:
            with self.port_chain(
                port_chain={
                    'port_pair_groups': [pg1['port_pair_group']['id']]
                }
            ) as pc1, self.port_chain(
                port_chain={
                    'port_pair_groups': [pg2['port_pair_group']['id']]
                }
            ) as pc2:
                self._create_service_graph(self.fmt, {
                    'port_chains': {
                        pc1['port_chain']['id']: [
                            pc2['port_chain']['id'],
                            pc2['port_chain']['id']
                        ]
                    },
                    'name': 'abc',
                    'description': 'def'
                }, expected_res_status=400)

    def test_create_service_graph_with_already_used_pcs_fail(self):
        # this test will attempt to create a graph that maps
        # port-chains which have already been mapped to other graphs.
        with self.port_pair_group(
            port_pair_group={}, do_delete=False
        ) as pg1, self.port_pair_group(
            port_pair_group={}, do_delete=False
        ) as pg2, self.port_pair_group(
            port_pair_group={}, do_delete=False
        ) as pg3:
            with self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg1['port_pair_group']['id']]}
            ) as pc1, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg2['port_pair_group']['id']]}
            ) as pc2, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg3['port_pair_group']['id']]}
            ) as pc3:
                self._create_service_graph(self.fmt, {
                    'port_chains': {
                        pc1['port_chain']['id']: [
                            pc2['port_chain']['id']
                        ]
                    },
                    'name': 'abc',
                    'description': 'def'
                }, expected_res_status=201)
                self._create_service_graph(self.fmt, {
                    'port_chains': {
                        pc3['port_chain']['id']: [
                            pc1['port_chain']['id']
                        ]
                    },
                    'name': 'abc',
                    'description': 'def'
                }, expected_res_status=409)

    def test_create_service_graph_with_multiple_starts(self):
        # this test will create a graph with multiple starting chains (tails)
        with self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg1, self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg2, self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg3, self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg4:
            with self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg1['port_pair_group']['id']]}
            ) as pc1, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg2['port_pair_group']['id']]}
            ) as pc2, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg3['port_pair_group']['id']]}
            ) as pc3, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg4['port_pair_group']['id']]}
            ) as pc4:
                self._create_service_graph(self.fmt, {
                    'port_chains': {
                        pc1['port_chain']['id']: [pc2['port_chain']['id']],
                        pc3['port_chain']['id']: [pc4['port_chain']['id']],
                        pc4['port_chain']['id']: [pc2['port_chain']['id']]
                    },
                    'name': 'abc',
                    'description': 'def'
                }, expected_res_status=201)

    def _test_create_service_graph_single_branching_two_fcs_each(
            self, fc1_dict, fc2_dict, fc3_dict, fc4_dict, expected_res_status
    ):
        with self.flow_classifier(
            flow_classifier=fc1_dict, do_delete=False
        ) as fc1, self.flow_classifier(
            flow_classifier=fc2_dict, do_delete=False
        ) as fc2, self.flow_classifier(
            flow_classifier=fc3_dict, do_delete=False
        ) as fc3, self.flow_classifier(
            flow_classifier=fc4_dict, do_delete=False
        ) as fc4:
            with self.port_pair_group(
                port_pair_group={}, do_delete=False
            ) as pg1, self.port_pair_group(
                port_pair_group={}, do_delete=False
            ) as pg2, self.port_pair_group(
                port_pair_group={}, do_delete=False
            ) as pg3:
                with self.port_chain(
                    port_chain={
                        'port_pair_groups': [pg1['port_pair_group']['id']]
                    },
                    do_delete=False
                ) as pc1, self.port_chain(
                    port_chain={
                        'port_pair_groups': [pg2['port_pair_group']['id']],
                        'flow_classifiers': [
                            fc1['flow_classifier']['id'],
                            fc2['flow_classifier']['id']
                        ]
                    },
                    do_delete=False
                ) as pc2, self.port_chain(
                    port_chain={
                        'port_pair_groups': [pg3['port_pair_group']['id']],
                        'flow_classifiers': [
                            fc3['flow_classifier']['id'],
                            fc4['flow_classifier']['id']
                        ]
                    },
                    do_delete=False
                ) as pc3:
                    self._create_service_graph(self.fmt, {
                        'port_chains': {
                            pc1['port_chain']['id']: [
                                pc2['port_chain']['id'],
                                pc3['port_chain']['id']
                            ]
                        },
                        'name': 'abc',
                        'description': 'def'
                    }, expected_res_status=expected_res_status)

    def test_create_service_graph_unambiguous_branch(self):
        # this test will create a graph where 1 port chain will act
        # as a dependency to 2 other port chains, using different
        # classifications for the dependent chains, which must succeed.
        with self.port(
            name='test1', do_delete=False
        ) as port1, self.port(
            name='test2', do_delete=False
        ) as port2, self.port(
            name='test3', do_delete=False
        ) as port3, self.port(
            name='test4', do_delete=False
        ) as port4:
            fc1_dict = {
                'name': 'fc1',
                'ethertype': 'IPv4',
                'protocol': 'tcp',
                'logical_source_port': port1['port']['id']
            }
            fc2_dict = {
                'name': 'fc2',
                'ethertype': 'IPv6',
                'protocol': 'tcp',
                'logical_source_port': port2['port']['id']
            }
            fc3_dict = {
                'name': 'fc3',
                'ethertype': 'IPv4',
                'protocol': 'udp',
                'logical_source_port': port3['port']['id']
            }
            fc4_dict = {
                'name': 'fc4',
                'ethertype': 'IPv6',
                'protocol': 'udp',
                'logical_source_port': port4['port']['id']
            }
            self._test_create_service_graph_single_branching_two_fcs_each(
                fc1_dict, fc2_dict, fc3_dict, fc4_dict,
                expected_res_status=201)

    def test_create_service_graph_with_direct_loop_fail(self):
        # this test will attempt to create a graph where there is a direct
        # loop, i.e. a chain linked to itself - specifically pc2->pc2.
        with self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg1, self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg2:
            with self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg1['port_pair_group']['id']]}
            ) as pc1, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg2['port_pair_group']['id']]}
            ) as pc2:
                self._create_service_graph(self.fmt, {
                    'port_chains': {
                        pc1['port_chain']['id']: [pc2['port_chain']['id']],
                        pc2['port_chain']['id']: [pc2['port_chain']['id']]
                    },
                    'name': 'abc',
                    'description': 'def'
                }, expected_res_status=400)

    def test_create_service_graph_with_indirect_loop_fail(self):
        # this test will attempt to create a graph where there is an indirect
        # loop, i.e. a chain is linked to a chain providing a path back to
        # the first chain again - specifically pc2->pc3->pc4->pc2.
        with self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg1, self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg2, self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg3, self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg4, self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg5:
            with self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg1['port_pair_group']['id']]}
            ) as pc1, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg2['port_pair_group']['id']]}
            ) as pc2, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg3['port_pair_group']['id']]}
            ) as pc3, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg4['port_pair_group']['id']]}
            ) as pc4, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg5['port_pair_group']['id']]}
            ) as pc5:
                self._create_service_graph(self.fmt, {
                    'port_chains': {
                        pc1['port_chain']['id']: [pc2['port_chain']['id']],
                        pc2['port_chain']['id']: [pc3['port_chain']['id']],
                        pc3['port_chain']['id']: [pc4['port_chain']['id']],
                        pc4['port_chain']['id']: [
                            pc2['port_chain']['id'],
                            pc5['port_chain']['id']
                        ]
                    },
                    'name': 'abc',
                    'description': 'def'
                }, expected_res_status=400)

    def test_create_service_graph_with_inexistent_port_chains(self):
        # this test will attempt to create a graph where one
        # of the referenced port chains do not exist, and fail.
        with self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg1, self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg2, self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg3:
            with self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg1['port_pair_group']['id']]}
            ) as pc1, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg2['port_pair_group']['id']]}
            ) as pc2, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg3['port_pair_group']['id']]}
            ) as pc3:
                self._create_service_graph(self.fmt, {
                    'port_chains': {
                        pc1['port_chain']['id']: [pc2['port_chain']['id']],
                        pc2['port_chain']['id']: [
                            pc3['port_chain']['id'],
                            uuidutils.generate_uuid()
                        ]
                    },
                    'name': 'abc',
                    'description': 'def'
                }, expected_res_status=404)

    def test_create_service_graph_with_joining_branches(self):
        # this test will create a graph that including "joining" branches, i.e.
        # a set of at least 2 branches that will be linked to the same next
        # port chain, thus joining traffic at that point.
        with self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg1, self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg2, self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg3, self.port_pair_group(
                port_pair_group={}, do_delete=False
        ) as pg4:
            with self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg1['port_pair_group']['id']]}
            ) as pc1, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg2['port_pair_group']['id']]}
            ) as pc2, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg3['port_pair_group']['id']]}
            ) as pc3, self.port_chain(
                do_delete=False,
                port_chain={'port_pair_groups': [pg4['port_pair_group']['id']]}
            ) as pc4:
                self._create_service_graph(self.fmt, {
                    'port_chains': {
                        pc1['port_chain']['id']: [pc2['port_chain']['id']],
                        pc2['port_chain']['id']: [
                            pc3['port_chain']['id'], pc4['port_chain']['id']
                        ],
                        pc3['port_chain']['id']: [pc4['port_chain']['id']]
                    },
                    'name': 'abc',
                    'description': 'def'
                }, expected_res_status=201)

    def test_update_service_graph(self):
        with self.port_pair_group(
            port_pair_group={}
        ) as pg1, self.port_pair_group(
            port_pair_group={}
        ) as pg2:
            with self.port_chain(
                port_chain={'port_pair_groups': [pg1['port_pair_group']['id']]}
            ) as pc1, self.port_chain(
                port_chain={'port_pair_groups': [pg2['port_pair_group']['id']]}
            ) as pc2:
                with self.service_graph(service_graph={
                    'name': 'test1',
                    'port_chains': {
                        pc1['port_chain']['id']: [pc2['port_chain']['id']]
                    }
                }) as graph:
                    updates = {
                        'name': 'test2',
                        'description': 'desc2'
                    }
                    req = self.new_update_request(
                        'service_graphs', {'service_graph': updates},
                        graph['service_graph']['id']
                    )
                    res = self.deserialize(
                        self.fmt,
                        req.get_response(self.ext_api)
                    )
                    expected = graph['service_graph']
                    expected.update(updates)
                    for k, v in expected.items():
                        self.assertEqual(res['service_graph'][k], v)
                    req = self.new_show_request(
                        'service_graphs', graph['service_graph']['id']
                    )
                    res = self.deserialize(
                        self.fmt, req.get_response(self.ext_api)
                    )
                    for k, v in expected.items():
                        self.assertEqual(res['service_graph'][k], v)

    def test_delete_service_graph(self):
        with self.port_pair_group(
            port_pair_group={}
        ) as pg1, self.port_pair_group(
            port_pair_group={}
        ) as pg2:
            with self.port_chain(
                port_chain={
                    'port_pair_groups': [pg1['port_pair_group']['id']]},
            ) as pc1, self.port_chain(
                port_chain={
                    'port_pair_groups': [pg2['port_pair_group']['id']]},
            ) as pc2:
                with self.service_graph(service_graph={
                    'name': 'test1',
                    'port_chains': {
                        pc1['port_chain']['id']: [pc2['port_chain']['id']]
                    }
                }, do_delete=False) as graph:
                    req = self.new_delete_request(
                        'service_graphs', graph['service_graph']['id']
                    )
                    res = req.get_response(self.ext_api)
                    self.assertEqual(204, res.status_int)
                    req = self.new_show_request(
                        'service_graphs', graph['service_graph']['id']
                    )
                    res = req.get_response(self.ext_api)
                    self.assertEqual(404, res.status_int)
                    req = self.new_show_request(
                        'port_chains', pc1['port_chain']['id']
                    )
                    res = req.get_response(self.ext_api)
                    self.assertEqual(200, res.status_int)
                    req = self.new_show_request(
                        'port_chains', pc2['port_chain']['id']
                    )
                    res = req.get_response(self.ext_api)
                    self.assertEqual(200, res.status_int)



# Copyright 2016 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.tests.functional.db import test_models


class TestDBCreation(test_models.TestDBCreation):
    def test_head_creation(self):
        super(TestDBCreation, self).test_head_creation()
        from networking_sfc.db.migration.models import head
        self._test_creation(head)



# Copyright 2016 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib import exceptions as lib_exc
from tempest.lib.services.network import base


class FlowClassifierClient(base.BaseNetworkClient):

    def create_flowclassifier(self, **kwargs):
        uri = '/sfc/flow_classifiers'
        post_data = {'flow_classifier': kwargs}
        return self.create_resource(uri, post_data)

    def update_flowclassifier(self, flowclassifier_id, **kwargs):
        uri = '/sfc/flow_classifiers/%s' % flowclassifier_id
        post_data = {'flow_classifier': kwargs}
        return self.update_resource(uri, post_data)

    def show_flowclassifier(self, flowclassifier_id, **fields):
        uri = '/sfc/flow_classifiers/%s' % flowclassifier_id
        return self.show_resource(uri, **fields)

    def delete_flowclassifier(self, flowclassifier_id):
        uri = '/sfc/flow_classifiers/%s' % flowclassifier_id
        return self.delete_resource(uri)

    def list_flowclassifiers(self, **filters):
        uri = '/sfc/flow_classifiers'
        return self.list_resources(uri, **filters)

    def is_resource_deleted(self, id):
        try:
            self.show_flowclassifier(id)
        except lib_exc.NotFound:
            return True
        return False

    @property
    def resource_type(self):
        """Returns the primary type of resource this client works with."""
        return 'flow_classifier'



# Copyright 2015 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from mock import Mock
from stevedore.extension import Extension

from neutron.tests import base

from networking_sfc.services.flowclassifier.common import exceptions as fc_exc
from networking_sfc.services.flowclassifier.driver_manager \
    import FlowClassifierDriverManager


class DriverManagerTestCase(base.BaseTestCase):
    def setUp(self):
        super(DriverManagerTestCase, self).setUp()

    def test_initialize_called(self):
        driver1 = Extension('mock_driver1', Mock(), None,
                            Mock(native_bulk_support=True))
        driver2 = Extension('mock_driver2', Mock(), None,
                            Mock(native_bulk_support=True))
        manager = FlowClassifierDriverManager.make_test_instance([driver1,
                                                                  driver2])
        manager.initialize()
        driver1.obj.initialize.assert_called_once_with()
        driver2.obj.initialize.assert_called_once_with()

    def _test_method_called(self, method_name):
        driver1 = Extension('mock_driver1', Mock(), None,
                            Mock(native_bulk_support=True))
        driver2 = Extension('mock_driver2', Mock(), None,
                            Mock(native_bulk_support=True))
        manager = FlowClassifierDriverManager.make_test_instance([driver1,
                                                                  driver2])
        mocked_context = Mock()
        getattr(manager, method_name)(mocked_context)
        getattr(driver1.obj, method_name).assert_called_once_with(
            mocked_context)
        getattr(driver2.obj, method_name).assert_called_once_with(
            mocked_context)

    def _test_method_exception(self, method_name,
                               expected_exc=fc_exc.FlowClassifierDriverError):
        driver = Extension('mock_driver', Mock(), None,
                           Mock(native_bulk_support=True))
        mock_method = Mock(side_effect=fc_exc.FlowClassifierException)
        setattr(driver.obj, method_name, mock_method)
        manager = FlowClassifierDriverManager.make_test_instance([driver])
        mocked_context = Mock()
        self.assertRaises(expected_exc,
                          getattr(manager, method_name),
                          mocked_context)

    def test_create_flow_classifier_precommit_called(self):
        self._test_method_called("create_flow_classifier_precommit")

    def test_create_flow_classifier_precommit_exception(self):
        self._test_method_exception("create_flow_classifier_precommit",
                                    fc_exc.FlowClassifierException)

    def test_create_flow_classifier_postcommit_called(self):
        self._test_method_called("create_flow_classifier_postcommit")

    def test_create_flow_classifier_postcommit_exception(self):
        self._test_method_exception("create_flow_classifier_postcommit")

    def test_update_flow_classifier_precommit_called(self):
        self._test_method_called("update_flow_classifier_precommit")

    def test_update_flow_classifier_precommit_exception(self):
        self._test_method_exception("update_flow_classifier_precommit")

    def test_update_flow_classifier_postcommit_called(self):
        self._test_method_called("update_flow_classifier_postcommit")

    def test_update_flow_classifier_postcommit_exception(self):
        self._test_method_exception("update_flow_classifier_postcommit")

    def test_delete_flow_classifier_called(self):
        self._test_method_called("delete_flow_classifier")

    def test_delete_flow_classifier_exception(self):
        self._test_method_exception("delete_flow_classifier")

    def test_delete_flow_classifier_precommit_called(self):
        self._test_method_called("delete_flow_classifier_precommit")

    def test_delete_flow_classifier_precommit_exception(self):
        self._test_method_exception("delete_flow_classifier_precommit")

    def test_delete_flow_classifier_postcommit_called(self):
        self._test_method_called("delete_flow_classifier_postcommit")

    def test_delete_flow_classifier_postcommit_exception(self):
        self._test_method_exception("delete_flow_classifier_postcommit")



# Copyright 2016 Futurewei. All rights reserved.
# Copyright 2017 Intel Corporation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import socket

import netaddr
from tempest.api.network import base
from tempest.lib.common.utils import data_utils
from tempest.lib import exceptions as lib_exc
from tempest import test

from networking_sfc.tests.tempest_plugin.tests import flowclassifier_client
from networking_sfc.tests.tempest_plugin.tests import sfc_client


class BaseFlowClassifierTest(
    flowclassifier_client.FlowClassifierClientMixin,
    base.BaseAdminNetworkTest
):
    @classmethod
    def resource_setup(cls):
        super(BaseFlowClassifierTest, cls).resource_setup()
        if not test.is_extension_enabled('flow_classifier', 'network'):
            msg = "FlowClassifier Extension not enabled."
            raise cls.skipException(msg)
        cls.network = cls.create_network()
        cls.subnet = cls.create_subnet(cls.network)
        cls.host_id = socket.gethostname()

    @classmethod
    def resource_cleanup(cls):
        if not test.is_extension_enabled('flow_classifier', 'network'):
            msg = "FlowClassifier Extension not enabled."
            raise cls.skipException(msg)
        super(BaseFlowClassifierTest, cls).resource_cleanup()

    @classmethod
    def _create_port(cls, network, **kwargs):
        body = cls.admin_ports_client.create_port(
            network_id=network['id'],
            **kwargs)
        port = body['port']
        return port

    def _try_create_flowclassifier(self, **kwargs):
        if 'logical_source_port' not in kwargs:
            port_kwargs = {"binding:host_id": self.host_id}
            port = self._create_port(network=self.network, **port_kwargs)
            self.addCleanup(self._try_delete_port, port['id'])
            kwargs['logical_source_port'] = port['id']
            if 'source_ip_prefix' not in kwargs:
                port_ip_prefix = str(netaddr.IPNetwork(
                    port['fixed_ips'][0]['ip_address']))
                kwargs['source_ip_prefix'] = port_ip_prefix
        fc = self.create_flowclassifier(**kwargs)
        self.addCleanup(self._try_delete_flowclassifier, fc['id'])
        return fc

    def _try_delete_port(self, port_id):
        try:
            self.admin_ports_client.delete_port(port_id)
        except lib_exc.NotFound:
            pass
        body = self.admin_ports_client.list_ports()
        ports_list = body['ports']
        self.assertNotIn(port_id, [n['id'] for n in ports_list])

    def _try_delete_flowclassifier(self, fc_id):
        # delete flowclassifier, if it exists
        try:
            self.flowclassifier_client.delete_flowclassifier(fc_id)
        # if flowclassifier is not found, this means it was deleted
        except lib_exc.NotFound:
            pass
        body = self.flowclassifier_client.list_flowclassifiers()
        fc_list = body['flow_classifiers']
        self.assertNotIn(fc_id, [n['id'] for n in fc_list])


class BaseSfcTest(
    sfc_client.SfcClientMixin, BaseFlowClassifierTest
):
    @classmethod
    def resource_setup(cls):
        super(BaseSfcTest, cls).resource_setup()
        if not test.is_extension_enabled('sfc', 'network'):
            msg = "Sfc Extension not enabled."
            raise cls.skipException(msg)

    @classmethod
    def resource_cleanup(cls):
        if not test.is_extension_enabled('sfc', 'network'):
            msg = "Sfc Extension not enabled."
            raise cls.skipException(msg)
        super(BaseSfcTest, cls).resource_cleanup()

    def _try_create_port_pair(self, **kwargs):
        if 'ingress' not in kwargs or 'egress' not in 'kwargs':
            router = self.admin_routers_client.create_router(
                name=data_utils.rand_name('router-'))['router']
            self.addCleanup(
                self.admin_routers_client.delete_router, router['id'])
            port_kwargs = {"binding:host_id": self.host_id}
            port = self._create_port(
                network=self.network, **port_kwargs)
            self.addCleanup(self._try_delete_port, port['id'])
            self.admin_routers_client.add_router_interface(
                router['id'], port_id=port['id'])
            self.addCleanup(self.admin_routers_client.remove_router_interface,
                            router['id'],
                            port_id=port['id'])
            if 'ingress' not in kwargs:
                kwargs['ingress'] = port['id']
            if 'egress' not in kwargs:
                kwargs['egress'] = port['id']
        pp = self.create_port_pair(**kwargs)
        self.addCleanup(self._try_delete_port_pair, pp['id'])
        return pp

    def _try_delete_port_pair(self, pp_id):
        # delete port pair, if it exists
        try:
            self.portpair_client.delete_port_pair(pp_id)
        # if port pair is not found, this means it was deleted
        except lib_exc.NotFound:
            pass
        body = self.portpair_client.list_port_pairs()
        pp_list = body['port_pairs']
        self.assertNotIn(pp_id, [n['id'] for n in pp_list])

    def _try_create_port_pair_group(self, **kwargs):
        pg = self.create_port_pair_group(
            **kwargs)
        self.addCleanup(self._try_delete_port_pair_group, pg['id'])
        # self.pgs.append(pg)
        return pg

    def _try_delete_port_pair_group(self, pg_id):
        # delete port pair group, if it exists
        try:
            self.portpairgroup_client.delete_port_pair_group(pg_id)
        # if port pair group is not found, this means it was deleted
        except lib_exc.NotFound:
            pass
        body = self.portpairgroup_client.list_port_pair_groups()
        pg_list = body['port_pair_groups']
        self.assertNotIn(pg_id, [n['id'] for n in pg_list])

    def _try_create_port_chain(self, **kwargs):
        pc = self.create_port_chain(
            **kwargs)
        self.addCleanup(self._try_delete_port_chain, pc['id'])
        # self.pcs.append(pc)
        return pc

    def _try_delete_port_chain(self, pc_id):
        # delete port chain, if it exists
        try:
            self.portchain_client.delete_port_chain(pc_id)
        # if port chain is not found, this means it was deleted
        except lib_exc.NotFound:
            pass
        body = self.portchain_client.list_port_chains()
        pc_list = body['port_chains']
        self.assertNotIn(pc_id, [n['id'] for n in pc_list])

    def _try_create_service_graph(self, **kwargs):
        graph = self.create_service_graph(
            **kwargs)
        self.addCleanup(self._try_delete_service_graph, graph['id'])
        return graph

    def _try_delete_service_graph(self, graph_id):
        # delete Service Graph, if it exists
        try:
            self.sfcgraph_client.delete_service_graph(graph_id)
        # if Service Graph is not found, this means it was deleted
        except lib_exc.NotFound:
            pass
        body = self.sfcgraph_client.list_service_graphs()
        graph_list = body['service_graphs']
        self.assertNotIn(graph_id, [n['id'] for n in graph_list])



# Copyright 2016 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from tempest import config
from tempest.lib import exceptions as lib_exc

from networking_sfc.tests.tempest_plugin.tests import flowclassifier_client
from networking_sfc.tests.tempest_plugin.tests.scenario import manager
from networking_sfc.tests.tempest_plugin.tests import sfc_client

CONF = config.CONF
LOG = logging.getLogger(__name__)


class SfcScenarioTest(
    flowclassifier_client.FlowClassifierClientMixin,
    sfc_client.SfcClientMixin,
    manager.NetworkScenarioTest
):
    def _check_connectivity(
        self, source_ip, destination_ip, routes=None,
        username=None, private_key=None
    ):
        msg = "ip address %r is reachable" % source_ip
        ok = self.ping_ip_address(source_ip, should_succeed=True)
        self.assertTrue(ok, msg=msg)
        client = self.get_remote_client(
            source_ip, username=username, private_key=private_key)
        cmd = 'traceroute -n -I %s' % destination_ip
        LOG.debug('exec command on %s: %s', source_ip, cmd)
        try:
            result = client.exec_command(cmd)
            LOG.debug(
                'traceroute from %s to %s:\n%s',
                source_ip, destination_ip, result)
            lines = result.split('\n')
            lines = [line for line in lines if line]
            lines = lines[1:-1]
            if len(lines) != len(routes):
                LOG.error('length mismatch:\n%s\nvs\n%s', lines, routes)
                self.assertEqual(len(lines), len(routes))
            for line, route_list in zip(lines, routes):
                found = any([route in line for route in route_list])
                if not found:
                    LOG.error('did not found any route %s in %s',
                              route_list, line)
                    self.assertTrue(found)
        except lib_exc.SSHExecCommandFailed as e:
            LOG.exception(e)
            raise
        except lib_exc.SSHTimeout as e:
            LOG.exception(e)
            raise



# Copyright 2015 Huawei Technologies India Pvt. Ltd.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import sys

import mock

from neutronclient import shell
from neutronclient.tests.unit import test_cli20

from networking_sfc.cli import port_pair_group as pg

from oslo_utils import uuidutils

pp1 = uuidutils.generate_uuid()
pp2 = uuidutils.generate_uuid()
pp3 = uuidutils.generate_uuid()
pp4 = uuidutils.generate_uuid()


class CLITestV20PortGroupExtensionJSON(test_cli20.CLITestV20Base):
    def setUp(self):
        super(CLITestV20PortGroupExtensionJSON, self).setUp()
        self._mock_extension_loading()
        self.register_non_admin_status_resource('port_pair_group')

    def _create_patch(self, name, func=None):
        patcher = mock.patch(name)
        thing = patcher.start()
        self.addCleanup(patcher.stop)
        return thing

    def _mock_extension_loading(self):
        ext_pkg = 'neutronclient.common.extension'
        port_pair_group = self._create_patch(ext_pkg +
                                             '._discover_via_entry_points')
        port_pair_group.return_value = [("port_pair_group", pg)]
        return port_pair_group

    def test_ext_cmd_loaded(self):
        neutron_shell = shell.NeutronShell('2.0')
        ext_cmd = {'port-pair-group-list': pg.PortPairGroupList,
                   'port-pair-group-create': pg.PortPairGroupCreate,
                   'port-pair-group-update': pg.PortPairGroupUpdate,
                   'port-pair-group-delete': pg.PortPairGroupDelete,
                   'port-pair-group-show': pg.PortPairGroupShow}
        for cmd_name, cmd_class in ext_cmd.items():
            found = neutron_shell.command_manager.find_command([cmd_name])
            self.assertEqual(cmd_class, found[0])

    def test_create_port_pair_group_with_mandatory_args(self):
        """Create port_pair_group: myname."""
        resource = 'port_pair_group'
        cmd = pg.PortPairGroupCreate(test_cli20.MyApp(sys.stdout), None)
        name = 'myname'
        myid = 'myid'
        args = [name, '--port-pair', pp1]
        position_names = ['name', 'port_pairs']
        position_values = [name, [pp1]]
        self._test_create_resource(resource, cmd, name, myid, args,
                                   position_names, position_values)

    def test_create_port_pair_group_with_multi_port_pairs(self):
        """Create port_pair_group: myname with multiple port pairs"""
        resource = 'port_pair_group'
        cmd = pg.PortPairGroupCreate(test_cli20.MyApp(sys.stdout), None)
        name = 'myname'
        myid = 'myid'
        args = [name, '--port-pair', pp1, '--port-pair', pp2]
        position_names = ['name', 'port_pairs']
        position_values = [name, [pp1, pp2]]
        self._test_create_resource(resource, cmd, name, myid, args,
                                   position_names, position_values)

    def test_create_port_pair_group_with_lb_fields_param(self):
        """Create port_pair_group: myname with lb_fields parameter"""
        resource = 'port_pair_group'
        cmd = pg.PortPairGroupCreate(test_cli20.MyApp(sys.stdout),
                                     None)
        name = 'myname'
        myid = 'myid'
        ppg_param = 'lb_fields=ip_src&ip_dst'
        ppg_exp = {"lb_fields": ["ip_src", "ip_dst"]}
        args = [name, '--port-pair', pp1,
                '--port-pair-group-parameters', ppg_param]
        position_names = ['name', 'port_pairs',
                          'port_pair_group_parameters']
        position_values = [name, [pp1], ppg_exp]
        self._test_create_resource(resource, cmd, name, myid, args,
                                   position_names, position_values)

    def test_create_port_pair_group_with_ppg_n_tuple_mapping_param(self):
        """Create port_pair_group: myname with ppg_n_tuple_mapping parameter"""
        resource = 'port_pair_group'
        cmd = pg.PortPairGroupCreate(test_cli20.MyApp(sys.stdout),
                                     None)
        name = 'myname'
        myid = 'myid'
        ppg_param = ('ppg_n_tuple_mapping=source_ip_prefix_ingress=None'
                     '&source_ip_prefix_egress=None')
        ppg_exp = {
            'ppg_n_tuple_mapping': {
                'ingress_n_tuple': {'source_ip_prefix': 'None'},
                'egress_n_tuple': {'source_ip_prefix': 'None'}}}
        args = [name, '--port-pair', pp1,
                '--port-pair-group-parameters', ppg_param]
        position_names = ['name', 'port_pairs',
                          'port_pair_group_parameters']
        position_values = [name, [pp1], ppg_exp]
        self._test_create_resource(resource, cmd, name, myid, args,
                                   position_names, position_values)

    def test_delete_port_pair_group(self):
        """Delete port_pair_group: myid."""
        resource = 'port_pair_group'
        cmd = pg.PortPairGroupDelete(test_cli20.MyApp(sys.stdout), None)
        myid = 'myid'
        args = [myid]
        self._test_delete_resource(resource, cmd, myid, args)

    def test_update_port_group_only_port_pair(self):
        """Update port_pair_group"""
        resource = 'port_pair_group'
        cmd = pg.PortPairGroupUpdate(test_cli20.MyApp(sys.stdout), None)
        myid = 'myid'
        args = [myid, '--port-pair', pp1,
                '--port-pair', pp2]
        updatefields = {'port_pairs': [pp1, pp2]}
        self._test_update_resource(resource, cmd, myid, args, updatefields)

    def test_update_port_group_with_all_desc(self):
        """Update port_pair_group and description"""
        resource = 'port_pair_group'
        cmd = pg.PortPairGroupUpdate(test_cli20.MyApp(sys.stdout), None)
        myid = 'myid'
        args = [myid, '--port-pair', pp1, '--port-pair', pp2,
                '--description', 'my_port_pair_group']
        updatefields = {'port_pairs': [pp1, pp2],
                        'description': 'my_port_pair_group'}
        self._test_update_resource(resource, cmd, myid, args, updatefields)

    def test_list_port_pair_group(self):
        """List port_pair_group."""
        resources = 'port_pair_groups'
        cmd = pg.PortPairGroupList(test_cli20.MyApp(sys.stdout), None)
        self._test_list_resources(resources, cmd, True)

    def test_list_port_pair_group_limit(self):
        """size (1000) limited list: port-pair-group -P."""
        resources = "port_pair_groups"
        cmd = pg.PortPairGroupList(test_cli20.MyApp(sys.stdout), None)
        self._test_list_resources(resources, cmd, page_size=1000)

    def test_list_port_group_sort(self):
        """List port_pair_group: --sort-key name --sort-key id --sort-key asc

        --sort-key desc
        """
        resources = "port_pair_groups"
        cmd = pg.PortPairGroupList(test_cli20.MyApp(sys.stdout), None)
        self._test_list_resources(resources, cmd,
                                  sort_key=["name", "id"],
                                  sort_dir=["asc", "desc"])

    def test_show_port_group(self):
        """Show port-chain: --fields id --fields name myid."""
        resource = 'port_pair_group'
        cmd = pg.PortPairGroupShow(test_cli20.MyApp(sys.stdout), None)
        args = ['--fields', 'id', '--fields', 'name', self.test_id]
        self._test_show_resource(resource, cmd, self.test_id,
                                 args, ['id', 'name'])
